;	RCS Header $Id: FLO0832.ASM,v 1.2 2000/02/22 22:15:27 krt Exp $

;	$Revision: 1.2 $

	include <P17C44.INC>		; General 17CXXX definitions

	include <CMATH17.INC>		; Math library definitions

;**********************************************************************************************
;**********************************************************************************************

;       8 bit unsigned integer to 32 bit float conversion

;       Input:  8 bit unsigned integer right justified in AARGB0

;       Use:    CALL    FLO0832U

;       Output: 32 bit floating point number in AEXP, AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  FLOAT( AARG )

;	Testing on [0x00,0xFF] from 100000 trials:

;		min	max	mean
;	Timing:	18	35	27.8	clks

;		min	max	mean	rms
;	Error:	0x00	0x00	0.00	0.00	nsb

;       PM: 7+29 = 36					DM: 7

;----------------------------------------------------------------------------------------------

PROG	CODE

		EXTERN		RES032

FLO0832U
		GLOBAL		FLO0832U

		MOVLW		D'7'+EXPBIAS		; initialize exponent and add bias
		MOVWF		AEXP
		CLRF 		SIGN,F			; clear sign
		CLRF 		AARGB1,F		; clear remaining bytes
		CLRF 		AARGB2,F
		CLRF 		AARGB3,F		; clear extended byte
		GOTO 		NRM1632

;**********************************************************************************************

;       8 bit signed integer to 32 bit float conversion

;       Input:  8 bit signed integer right justified in AARGB0

;       Use:    CALL    FLO0832S

;       Output: 32 bit floating point number in AEXP, AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  FLOAT( AARG )

;	Testing on [-0x7F,0x7F] from 100000 trials:

;		min	max	mean
;	Timing:	20	38	33.0	clks

;		min	max	mean	rms
;	Error:	0x00	0x00	0.00	0.00	nsb

;       PM: 11+29 = 40					DM: 7

;----------------------------------------------------------------------------------------------

FLO0832S
		GLOBAL		FLO0832S

		MOVLW		D'7'+EXPBIAS		; initialize exponent and add bias
		MOVWF		AEXP
		MOVFP		AARGB0,WREG
		MOVWF		SIGN			; save sign in SIGN
		CLRF 		AARGB1,F		; clear remaining bytes
		CLRF 		AARGB2,F
		CLRF 		AARGB3,W		; clear extended byte
		BTFSS		AARGB0,MSB		; test sign
		GOTO 		NRM1632
		COMF 		AARGB0,F		; if < 0, negate
		INCF 		AARGB0,F

;**********************************************************************************************

;       Normalization routine

;       Input:  16 bit unnormalized floating point number in AEXP, AARGB0,
;               with AARGB1 = AARGB2 = 0 and sign in SIGN,MSB

;       Use:    CALL    NRM1632

;       Output: 32 bit normalized floating point number in AEXP, AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  NORMALIZE( AARG )

;       Max Timing:     9+13+8 = 30 clks

;       Min Timing:     4+7 = 11 clks			AARG = 0
;                       3+17 = 20 clks

;       PM: 29                                         DM: 6

;----------------------------------------------------------------------------------------------

NRM1632
		CLRF  		TEMP,W			; clear exponent decrement
		CPFSGT		AARGB0			; test if highbyte=0
		GOTO  		RES032

TNIB1632
		MOVLW 		0xF0                    ; test if highnibble=0
		ANDWF 		AARGB0,W
		TSTFSZ		WREG
		GOTO  		NORM1632
		SWAPF 		AARGB0,F                ; if so, shift 4 bits

		BSF   		TEMP,2                  ; increase decrement by 4

NORM1632
		BCF  		_C                      ; clear carry bit

		BTFSC		AARGB0,MSB		; if MSB=1, normalization done
		GOTO 		TNORMUN1632
		RLCF 		AARGB0,F                ; otherwise, shift left and 
		INCF 		TEMP,F			; increment decrement
		BTFSC		AARGB0,MSB
		GOTO 		TNORMUN1632
		RLCF 		AARGB0,F
		INCF 		TEMP,F
		BTFSC		AARGB0,MSB		; since highnibble != 0, at most
		GOTO 		TNORMUN1632             ; 3 left shifts are required
		RLCF 		AARGB0,F
		INCF 		TEMP,F

TNORMUN1632
		MOVFP		TEMP,WREG

;	For use in FLO0832 only, underflow cannot occur. To normalize arbitrary 16 bit
;	denormalized numbers, reinstate this underflow test.

;		CPFSGT		AEXP			; if EXP <= decrement in TEMP,floating
;		GOTO  		SETFUN32		; point underflow has occured

		SUBWF		AEXP,F			; compute EXP

		BTFSS		SIGN,MSB		; rounding unnecessary
		BCF  		AARGB0,MSB		; clear explicit MSB if positive
		RETLW		0x00

		END
