;	RCS Header $Id: FLO1632.ASM,v 1.2 2000/02/22 22:15:27 krt Exp $

;	$Revision: 1.2 $

	include <P17C44.INC>		; General 17CXXX definitions

	include <CMATH17.INC>		; Math library definitions

;**********************************************************************************************
;**********************************************************************************************

;       16 bit unsigned integer to 32 bit float conversion

;       Input:  16 bit unsigned integer right justified in AARGB0,AARGB1

;       Use:    CALL    FLO1632U

;       Output: 32 bit floating point number in AEXP, AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  FLOAT( AARG )

;	Testing on [0x0000,0xFFFF] from 100000 trials:

;		min	max	mean
;	Timing:	27	46	30.9	clks

;		min	max	mean	rms
;	Error:	0x00	0x00	0.00	0.00	nsb

;       PM: 6+62 = 68					DM: 6

;----------------------------------------------------------------------------------------------

PROG	CODE

		EXTERN		RES032
		EXTERN		FIXSIGN32

FLO1632U
		GLOBAL		FLO1632U

		MOVLW		D'15'+EXPBIAS		; initialize exponent and add bias
		MOVWF		AEXP
		CLRF 		SIGN,F			; clear sign
		CLRF 		AARGB2,F		; clear remaining byte
		CLRF 		AARGB3,F		; clear extended byte
		GOTO 		NRM2432

;**********************************************************************************************

;       Integer to float conversion

;       Input:  16 bit 2's complement integer right justified in AARGB0,AARGB1

;       Use:    CALL    FLO1632S

;       Output: 32 bit floating point number in AEXP, AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  FLOAT( AARG )

;	Testing on [-7FFF,0x7FFF] from 100000 trials:

;		min	max	mean
;	Timing:	27	51	38.3	clks

;		min	max	mean	rms
;	Error:	0x00	0x00	0.00	0.00	nsb

;       PM: 12+62 = 74                                DM: 6

;----------------------------------------------------------------------------------------------

FLO1632S
		GLOBAL		FLO1632S

		MOVLW 		D'15'+EXPBIAS		; initialize exponent and add bias
		MOVWF 		AEXP
		MOVFP 		AARGB0,WREG
		MOVWF 		SIGN			; save sign in SIGN
		CLRF  		AARGB2,F		; clear remaining byte
		CLRF  		AARGB3,W		; clear extended byte
		BTFSS 		AARGB0,MSB		; test sign
		GOTO  		NRM2432
		COMF  		AARGB1,F		; if < 0, negate
		COMF  		AARGB0,F
		INCF  		AARGB1,F
		ADDWFC		AARGB0,F

;**********************************************************************************************

;       Normalization routine

;       Input:  24 bit unnormalized floating point number in AEXP, AARGB0, AARGB1,
;               with AARGB2 = 0 and sign in SIGN,MSB

;       Use:    CALL    NRM2432

;       Output: 32 bit normalized floating point number in AEXP, AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  NORMALIZE( AARG )


;       Max Timing:     16+16+6+4 = 42 clks

;       Min Timing:     4+7+8 = 19 clks			AARG = 0
;                       4+15+6+4 = 29 clks

;       PM: 62						DM: 6

;----------------------------------------------------------------------------------------------

NRM2432
		CLRF  		TEMP,W			; clear exponent decrement
		CPFSGT		AARGB0			; test if highbyte=0
		GOTO  		NRM2432A

TNIB2432
		MOVLW 		0xF0                    ; test if highnibble=0
		ANDWF 		AARGB0,W
		TSTFSZ		WREG
		GOTO  		NORM2432
		SWAPF 		AARGB0,F                ; if so, shift 4 bits
		SWAPF 		AARGB1,W
		ANDLW 		0x0F
		ADDWF 		AARGB0,F

		SWAPF 		AARGB1,W
		ANDLW 		0xF0
		MOVPF 		WREG,AARGB1

		BSF   		TEMP,2                  ; increase decrement by 4

NORM2432
		BCF  		_C                      ; clear carry bit

		BTFSC		AARGB0,MSB		; if MSB=1, normalization done
		GOTO 		TNORMUN2432
		RLCF 		AARGB1,F                ; otherwise, shift left and 
		RLCF 		AARGB0,F                ; increment decrement
		INCF 		TEMP,F
		BTFSC		AARGB0,MSB
		GOTO 		TNORMUN2432
		RLCF 		AARGB1,F
		RLCF 		AARGB0,F
		INCF 		TEMP,F
		BTFSC		AARGB0,MSB		; since highnibble != 0, at most
		GOTO 		TNORMUN2432             ; 3 left shifts are required
		RLCF 		AARGB1,F
		RLCF 		AARGB0,F
		INCF 		TEMP,F

TNORMUN2432
		 MOVFP		TEMP,WREG

;	For use in FLO1632 only, underflow cannot occur. To normalize arbitrary 24 bit
;	denormalized numbers, reinstate this underflow test.

;		CPFSGT		AEXP			; if EXP <= decrement in TEMP,floating
;		GOTO  		SETFUN32		; point underflow has occured

		SUBWF		AEXP,F			; compute EXP

		GOTO 		FIXSIGN32		; rounding unnecessary

NRM2432A
		MOVFP 		AARGB1,WREG
		MOVWF 		AARGB0			; shift 8 bits by move
		CLRF  		AARGB1,W
		BSF   		TEMP,3                  ; increase decrement by 8       
		CPFSGT		AARGB0			; test if highbyte=0
		GOTO  		RES032

TNIB2432A
		MOVLW 		0xF0                    ; test if highnibble=0
		ANDWF 		AARGB0,W
		TSTFSZ		WREG
		GOTO  		NORM2432A
		SWAPF 		AARGB0,F                ; if so, shift 4 bits

		BSF   		TEMP,2                  ; increase decrement by 4

NORM2432A
		BCF  		_C			; clear carry bit

		BTFSC		AARGB0,MSB		; if MSB=1, normalization done
		GOTO 		TNORMUN2432
		RLCF 		AARGB0,F                ; otherwise, shift left and 
		INCF 		TEMP,F			; increment decrement
		BTFSC		AARGB0,MSB
		GOTO 		TNORMUN2432
		RLCF 		AARGB0,F
		INCF 		TEMP,F
		BTFSC		AARGB0,MSB		; since highnibble != 0, at most
		GOTO 		TNORMUN2432             ; 3 left shifts are required
		RLCF 		AARGB0,F
		INCF 		TEMP,F
		GOTO 		TNORMUN2432

		END
