;	RCS Header $Id: FLO3232.ASM,v 1.2 2000/02/22 22:15:27 krt Exp $

;	$Revision: 1.2 $

	include <P17C44.INC>		; General 17CXXX definitions

	include <CMATH17.INC>		; Math library definitions

;**********************************************************************************************
;**********************************************************************************************

;       32 bit unsigned integer to 32 bit float conversion

;       Input:  32 bit unsigned integer right justified in AARGB0,AARGB1,
;               AARGB2, AARGB3

;       Use:    CALL    FLO3232U

;       Output: 32 bit floating point number in AEXP, AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  FLOAT( AARG )

;	Testing on [0x00000000,0xFFFFFFFF] from 100000 trials:

;		min	max	mean
;	Timing:	29	70	35.1	clks

;		min	max	mean	rms
;	Error:	0x00	0x00	0.00	0.00	nsb

;       PM: 14+178 = 192                                DM: 8

;----------------------------------------------------------------------------------------------

PROG	CODE

		EXTERN		SETFUN32
		EXTERN		FIXSIGN32
		EXTERN		SETFOV32
		EXTERN		RES032

FLO3232U
		GLOBAL		FLO3232U

		MOVLW		D'31'+EXPBIAS		; initialize exponent and add bias
		MOVWF		AEXP
		CLRF 		SIGN,F			; clear sign
		GOTO 		NRM4032

;**********************************************************************************************

;       Integer to float conversion

;       Input:  32 bit 2's complement integer right justified in AARGB0,AARGB1,
;               AARGB2, AARGB3

;       Use:    CALL    FLO3232

;       Output: 32 bit floating point number in AEXP, AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  FLOAT( AARG )

;	Testing on [MINNUM,MAXNUM] from 100000 trials:

;			min	max		mean
;	Timing:	14	2024	1834.7	clks

;			min		max		mean	rms
;	Error:	-0xA2	0x9A	2.20	29.18	nsb

;       PM: 14+178 = 192                                DM: 8

;----------------------------------------------------------------------------------------------

FLO3232S
		GLOBAL		FLO3232S

		MOVLW 		D'31'+EXPBIAS		; initialize exponent and add bias
		MOVWF 		AEXP
		MOVFP 		AARGB0,WREG
		MOVWF		SIGN		; save sign in SIGN
		BTFSS 		AARGB0,MSB		; test sign
		GOTO  		NRM4032
		CLRF  		WREG,F			; if < 0, negate, set MSB in SIGN
		COMF  		AARGB3,F
		COMF  		AARGB2,F
		COMF  		AARGB1,F
		COMF  		AARGB0,F
		INCF  		AARGB3,F
		ADDWFC		AARGB2,F
		ADDWFC		AARGB1,F
		ADDWFC		AARGB0,F

;**********************************************************************************************

;       Normalization routine

;       Input:  40 bit unnormalized floating point number in AEXP, AARGB0, AARGB1,
;               AARGB2, AARGB3 with sign in SIGN,MSB.

;       Use:    CALL    NRM4032

;       Output: 32 bit normalized floating point number in AEXP, AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  NORMALIZE( AARG )


;	Testing on [MINNUM,MAXNUM] from 100000 trials:

;			min	max		mean
;	Timing:	14	2024	1834.7	clks

;			min		max		mean	rms
;	Error:	-0xA2	0x9A	2.20	29.18	nsb

;       PM: 178                                         DM: 8

;----------------------------------------------------------------------------------------------

NRM4032
		GLOBAL		NRM4032

		MOVFP		AARGB0,WREG
		ANDLW		0xFF
		BTFSC		_Z
		GOTO 		NRM4032A

		ANDLW		0xF0
		BTFSC		_Z
		GOTO 		NRM4032MUL
NRM4032SFT
		CLRF 		WREG,F
		BTFSC		AARGB0,MSB		; if MSB=1, normalization done
		GOTO 		NRMRND4032
		BCF	 		_C
		RLCF 		AARGB3,F                ; otherwise, shift left and 
		RLCF 		AARGB2,F                ; increment decrement
		RLCF 		AARGB1,F
		RLCF 		AARGB0,F
		INCF 		WREG,F
		BTFSC		AARGB0,MSB
		GOTO 		TNORMUN4032
		RLCF 		AARGB3,F
		RLCF 		AARGB2,F
		RLCF 		AARGB1,F
		RLCF 		AARGB0,F
		INCF 		WREG,F
		BTFSC		AARGB0,MSB		; since highnibble != 0, at most
		GOTO 		TNORMUN4032		; 3 left shifts are required
		RLCF 		AARGB3,F
		RLCF 		AARGB2,F
		RLCF 		AARGB1,F
		RLCF 		AARGB0,F
		INCF 		WREG,F

TNORMUN4032
			  						; if EXP <= decrement in WREG,
		CPFSGT		AEXP			; floating point underflow has
		GOTO  		SETFUN32                ; occured
		SUBWF 		AEXP,F			; otherwise, compute EXP

NRMRND4032
		GLOBAL		NRMRND4032

		BTFSC 		FPFLAGS,RND		; is rounding enabled?
		BTFSS 		AARGB3,MSB		; is NSB > 0x80?
		GOTO  		FIXSIGN32
		BSF	  		_C			; set carry for rounding
		MOVLW 		0x80
		CPFSGT		AARGB3			; if NSB = 0x80, select even
		RRCF  		AARGB2,W		; using lsb in carry
		CLRF  		WREG,F
		ADDWFC		AARGB2,F
		ADDWFC		AARGB1,F
		ADDWFC		AARGB0,F

		BTFSS 		_C                      ; has rounding caused carryout?
		GOTO  		FIXSIGN32
		RRCF  		AARGB0,F                ; if so, right shift
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		INFSNZ		AEXP,F			; test for floating point overflow
		GOTO  		SETFOV32
		GOTO  		FIXSIGN32


NRM4032MUL
		MOVFP 		AARGB2,WREG
		MOVWF 		BARGB2
		MOVFP 		AARGB0,WREG
		MOVWF 		BARGB0

		MOVFP 		AARGB0,WREG
		ADDLW 		LOW (NRMTBL16)		; access table for F0
		MOVWF 		TBLPTRL
		MOVLW 		HIGH (NRMTBL16)
		CLRF  		TBLPTRH,F
		ADDWFC		TBLPTRH,F
		TABLRD		0,1,WREG
		TLRD  		1,WREG

		MULWF 		AARGB3
		MOVPF 		PRODL,AARGB3
		MOVPF 		PRODH,AARGB2
		MULWF 		AARGB1
		MOVPF 		PRODL,AARGB1
		MOVPF 		PRODH,AARGB0
		MULWF 		BARGB2
		MOVPF 		PRODL,WREG
		ADDWF 		AARGB2,F
		MOVPF 		PRODH,WREG
		ADDWF 		AARGB1,F
		TLRD  		1,WREG
		MULWF 		BARGB0
		MOVPF 		PRODL,WREG
		ADDWF 		AARGB0,F

		TLRD  		0,WREG
		GOTO  		TNORMUN4032

NRM4032A
		MOVFP		AARGB1,WREG
		MOVWF		AARGB0		; if so, shift 8 bits by move
		ANDLW		0xFF
		BTFSC		_Z
		GOTO 		NRM4032B
		MOVFP		AARGB2,WREG
		MOVWF		AARGB1
		MOVFP		AARGB3,WREG
		MOVWF		AARGB2
		CLRF 		AARGB3,F

		MOVFP		AARGB0,WREG
		ANDLW		0xF0
		BTFSC		_Z
		GOTO 		NRM4032MULA
NRM4032SFTA
		MOVLW		0x08
		BTFSC		AARGB0,MSB		; if MSB=1, normalization done
		GOTO 		TNORMUN4032
		BCF	 		_C
		RLCF 		AARGB2,F                ; increment decrement
		RLCF 		AARGB1,F
		RLCF 		AARGB0,F
		INCF 		WREG,F
		BTFSC		AARGB0,MSB
		GOTO 		TNORMUN4032
		RLCF 		AARGB2,F
		RLCF 		AARGB1,F
		RLCF 		AARGB0,F
		INCF 		WREG,F
		BTFSC		AARGB0,MSB		; since highnibble != 0, at most
		GOTO 		TNORMUN4032		; 3 left shifts are required
		RLCF 		AARGB2,F
		RLCF 		AARGB1,F
		RLCF 		AARGB0,F
		INCF 		WREG,F

		GOTO 		TNORMUN4032

NRM4032MULA
		MOVFP 		AARGB1,WREG
		MOVWF 		BARGB1

		MOVFP 		AARGB0,WREG
		ADDLW 		LOW (NRMTBL16)		; access table for F0
		MOVWF 		TBLPTRL
		MOVLW 		HIGH (NRMTBL16)
		CLRF  		TBLPTRH,F
		ADDWFC		TBLPTRH,F
		TABLRD		0,1,WREG
		TLRD  		1,WREG

		MULWF 		AARGB2
		MOVPF 		PRODL,AARGB2
		MOVPF 		PRODH,AARGB1
		MULWF 		AARGB0
		MOVPF 		PRODL,AARGB0
		MULWF 		BARGB1
		MOVPF 		PRODL,WREG
		ADDWF 		AARGB1,F
		MOVPF 		PRODH,WREG
		ADDWF 		AARGB0,F

		TLRD  		0,WREG
		ADDLW 		0x08
		GOTO  		TNORMUN4032

NRM4032B
		MOVFP		AARGB2,WREG
		MOVWF		AARGB0		; if so, shift 8 bits by move
		ANDLW		0xFF
		BTFSC		_Z
		GOTO 		NRM4032C
		MOVFP		AARGB3,WREG
		MOVWF		AARGB1
		CLRF 		AARGB2,F
		CLRF 		AARGB3,F

		MOVFP		AARGB0,WREG
		ANDLW		0xF0
		BTFSC		_Z
		GOTO 		NRM4032MULB
NRM4032SFTB
		MOVLW		0x10
		BTFSC		AARGB0,MSB		; if MSB=1, normalization done
		GOTO 		TNORMUN4032
		BCF	 		_C
		RLCF 		AARGB1,F
		RLCF 		AARGB0,F
		INCF 		WREG,F
		BTFSC		AARGB0,MSB
		GOTO 		TNORMUN4032
		RLCF 		AARGB1,F
		RLCF 		AARGB0,F
		INCF 		WREG,F
		BTFSC		AARGB0,MSB		; since highnibble != 0, at most
		GOTO 		TNORMUN4032		; 3 left shifts are required
		RLCF 		AARGB1,F
		RLCF 		AARGB0,F
		INCF 		WREG,F

		GOTO 		TNORMUN4032

NRM4032MULB
		MOVFP 		AARGB0,WREG
		MOVWF 		BARGB0

		MOVFP 		AARGB0,WREG
		ADDLW 		LOW (NRMTBL16)		; access table for F0
		MOVWF 		TBLPTRL
		MOVLW 		HIGH (NRMTBL16)
		CLRF  		TBLPTRH,F
		ADDWFC		TBLPTRH,F
		TABLRD		0,1,WREG
		TLRD  		1,WREG

		MULWF 		AARGB1
		MOVPF 		PRODL,AARGB1
		MOVPF 		PRODH,AARGB0
		MULWF 		BARGB0
		MOVPF 		PRODL,WREG
		ADDWF 		AARGB0,F

		TLRD  		0,WREG
		ADDLW 		0x10
		GOTO  		TNORMUN4032

NRM4032C
		MOVFP		AARGB3,WREG
		MOVWF		AARGB0		; if so, shift 8 bits by move
		ANDLW		0xFF
		BTFSC		_Z
		GOTO 		RES032
		CLRF 		AARGB1,F
		CLRF 		AARGB2,F
		CLRF 		AARGB3,F

		MOVFP		AARGB0,WREG
		ANDLW		0xF0
		BTFSC		_Z
		GOTO 		NRM4032MULC
NRM4032SFTC
		MOVLW		0x18
		BTFSC		AARGB0,MSB		; if MSB=1, normalization done
		GOTO 		TNORMUN4032
		BCF	 		_C
		RLCF 		AARGB0,F
		INCF 		WREG,F
		BTFSC		AARGB0,MSB
		GOTO 		TNORMUN4032
		RLCF 		AARGB0,F
		INCF 		WREG,F
		BTFSC		AARGB0,MSB		; since highnibble != 0, at most
		GOTO 		TNORMUN4032		; 3 left shifts are required
		RLCF 		AARGB0,F
		INCF 		WREG,F

		GOTO 		TNORMUN4032

NRM4032MULC
		MOVFP 		AARGB0,WREG
		ADDLW 		LOW (NRMTBL16)		; access table for F0
		MOVWF 		TBLPTRL
		MOVLW 		HIGH (NRMTBL16)
		CLRF  		TBLPTRH,F
		ADDWFC		TBLPTRH,F
		TABLRD		0,1,WREG
		TLRD  		1,WREG

		MULWF 		AARGB0
		MOVPF 		PRODL,AARGB0

		TLRD  		0,WREG
		ADDLW 		0x18
		GOTO  		TNORMUN4032

NRMTBL16
		GLOBAL		NRMTBL16

		DATA		0x0000
		DATA		0x8007
		DATA		0x4006
		DATA		0x4006
		DATA		0x2005
		DATA		0x2005
		DATA		0x2005
		DATA		0x2005
		DATA		0x1004
		DATA		0x1004
		DATA		0x1004
		DATA		0x1004
		DATA		0x1004
		DATA		0x1004
		DATA		0x1004
		DATA		0x1004

		END
