;   RCS Header $Id: FLOOR32.ASM,v 1.1 1999/01/08 19:56:36 GrosbaJ Exp $

;   $Revision: 1.1 $

;**********************************************************************************************
;**********************************************************************************************

;   Evaluate floor(x)

;   Input:  32 bit floating point number in AEXP, AARGB0, AARGB1, AARGB2

;   Use:    CALL    FLOOR32

;   Output: 32 bit floating point number in AEXP, AARGB0, AARGB1, AARGB2

;   Result: AARG  <--  FLOOR( AARG )

;   Testing on [-MAXNUM,MAXNUM] from 100000 trials:

;       min max mean
;   Timing: 32  49  37.2    clks

;       min max mean    rms
;   Error:  0x00    0x00    0.0 0.0 nsb

;----------------------------------------------------------------------------------------------

        #include    "cmath17.inc"

        EXTERN      SETFOV32

WREG        EQU     H'0A'
TBLPTRL     EQU     H'0D'
TBLPTRH     EQU     H'0E'

FLOOR32_CODE  CODE

FLOOR32
        GLOBAL      FLOOR32

        PAGESEL     FLOOR32
        BANKSEL     AARGB0

        CLRF        AARGB3,W        ; test for zero argument
        CPFSGT      AEXP
        RETLW       0x00

        MOVFP       AARGB0,WREG
        MOVWF       AARGB4          ; save mantissa
        MOVFP       AARGB1,WREG
        MOVWF       AARGB5
        MOVFP       AARGB2,WREG
        MOVWF       AARGB6

        MOVLW       EXPBIAS
        SUBWF       AEXP,W
        BTFSC       WREG,MSB
        GOTO        FLOOR32ZERO

        SUBLW       0x18-1
        MOVWF       TEMPB0          ; save number of zero bits in TEMPB0

        BTFSC       WREG,LSB+1+3        ; divide by eight
        GOTO        FLOOR32MASKH
        BTFSC       WREG,LSB+3
        GOTO        FLOOR32MASKM

FLOOR32MASKL
        CLRF        TBLPTRH,F

        MOVFP       TEMPB0,WREG     ; get remainder for mask pointer
        ANDLW       0x07

        ADDLW       LOW (FLOOR32MASKTABLE)
        MOVWF       TBLPTRL
        MOVLW       HIGH (FLOOR32MASKTABLE) ; access table for F0
        ADDWFC      TBLPTRH,F
        TABLRD      0,1,WREG
        TLRD        0,WREG

        ANDWF       AARGB2,F
        BTFSS       AARGB0,MSB      ; if negative, round down
        RETLW       0x00

        MOVWF       AARGB7
        MOVFP       AARGB6,WREG
        CPFSEQ      AARGB2
        GOTO        FLOOR32RNDL
        RETLW       0x00

FLOOR32RNDL
        COMF        AARGB7,W
        INCF        WREG,F
        ADDWF       AARGB2,F
        CLRF        WREG,F
        ADDWFC      AARGB1,F
        ADDWFC      AARGB0,F
        BTFSS       _C          ; has rounding caused carryout?
        RETLW       0x00
        RRCF        AARGB0,F
        RRCF        AARGB1,F
        RRCF        AARGB2,F
        INCFSZ      AEXP,F          ; check for overflow
        RETLW       0x00
        GOTO        SETFOV32

FLOOR32MASKM
        CLRF        TBLPTRH,F

        MOVFP       TEMPB0,WREG
        ANDLW       0x07

        ADDLW       LOW (FLOOR32MASKTABLE)
        MOVWF       TBLPTRL
        MOVLW       HIGH (FLOOR32MASKTABLE) ; access table for F0
        ADDWFC      TBLPTRH,F
        TABLRD      0,1,WREG
        TLRD        0,WREG

        ANDWF       AARGB1,F
        CLRF        AARGB2,F
        BTFSS       AARGB0,MSB      ; if negative, round down
        RETLW       0x00

        MOVWF       AARGB7
        MOVFP       AARGB6,WREG
        CPFSEQ      AARGB2
        GOTO        FLOOR32RNDM
        MOVFP       AARGB5,WREG
        CPFSEQ      AARGB1
        GOTO        FLOOR32RNDM
        RETLW       0x00

FLOOR32RNDM
        COMF        AARGB7,W
        INCF        WREG,F
        ADDWF       AARGB1,F
        CLRF        WREG,F
        ADDWFC      AARGB0,F
        BTFSS       _C          ; has rounding caused carryout?
        RETLW       0x00
        RRCF        AARGB0,F
        RRCF        AARGB1,F
        RRCF        AARGB2,F
        INCFSZ      AEXP,F          ; check for overflow
        RETLW       0x00
        GOTO        SETFOV32

FLOOR32MASKH
        CLRF        TBLPTRH,F

        MOVFP       TEMPB0,WREG
        ANDLW       0x07

        ADDLW       LOW (FLOOR32MASKTABLE)
        MOVWF       TBLPTRL
        MOVLW       HIGH (FLOOR32MASKTABLE) ; access table for F0
        ADDWFC      TBLPTRH,F
        TABLRD      0,1,WREG
        TLRD        0,WREG

        ANDWF       AARGB0,F
        CLRF        AARGB1,F
        CLRF        AARGB2,F
        BTFSS       AARGB0,MSB      ; if negative, round down
        RETLW       0x00

        MOVWF       AARGB7
        MOVFP       AARGB6,WREG
        CPFSEQ      AARGB2
        GOTO        FLOOR32RNDH
        MOVFP       AARGB5,WREG
        CPFSEQ      AARGB1
        GOTO        FLOOR32RNDH
        MOVFP       AARGB4,WREG
        CPFSEQ      AARGB0
        GOTO        FLOOR32RNDH
        RETLW       0x00

FLOOR32RNDH
        COMF        AARGB7,W
        INCF        WREG,F
        ADDWF       AARGB0,F
        BTFSS       _C          ; has rounding caused carryout?
        RETLW       0x00
        RRCF        AARGB0,F
        RRCF        AARGB1,F
        INCFSZ      AEXP,F          ; check for overflow
        RETLW       0x00
        GOTO        SETFOV32

FLOOR32ZERO
        BTFSC       AARGB0,MSB
        GOTO        FLOOR32MINUSONE
        CLRF        AEXP,F
        CLRF        AARGB0,F
        CLRF        AARGB1,F
        CLRF        AARGB2,F
        RETLW       0x00

FLOOR32MINUSONE
        MOVLW       0x7F
        MOVWF       AEXP
        MOVLW       0x80
        MOVWF       AARGB0
        CLRF        AARGB1,F
        CLRF        AARGB2,F
        RETLW       0x00

;----------------------------------------------------------------------------------------------

;   table for least significant byte requiring masking, using pointer from 
;   the remainder of the number of zero bits divided by eight.

FLOOR32MASKTABLE
        DATA        0xFF
        DATA        0xFE
        DATA        0xFC
        DATA        0xF8
        DATA        0xF0
        DATA        0xE0
        DATA        0xC0
        DATA        0x80
        DATA        0x00


        END
