;	RCS Header $Id: FPD32.ASM,v 1.2 2000/02/22 22:15:27 krt Exp $

;	$Revision: 1.2 $

	include <P17C44.INC>		; General 17CXXX definitions

	include <CMATH17.INC>		; Math library definitions

;**********************************************************************************************
;**********************************************************************************************

;       Floating Point Divide

;       Input:  32 bit floating point dividend in AEXP, AARGB0, AARGB1, AARGB2
;               32 bit floating point divisor in BEXP, BARGB0, BARGB1, BARGB2

;       Use:    CALL    FPD32

;       Output: 32 bit floating point quotient in AEXP, AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  AARG / BARG

;	Testing on [MINNUM,MAXNUM] from 100000 trials:

;			min	max		mean
;	Timing:	14	2024	1834.7	clks

;			min		max		mean	rms
;	Error:	-0xA2	0x9A	2.20	29.18	nsb

;       PM: 350+257 = 607			DM: 20

FAST	EQU		0
MEDIUM	EQU		1
SLOW	EQU		2

DIVIDE	EQU		FAST

;----------------------------------------------------------------------------------------------

FPD32SEEDF	macro

;       Timing:     34 clks

;	PM: 32+257 = 289

;	generation of F0 by interpolating between consecutive 16 bit approximations
;	to the reciprocal of BARG, with the top 8 explicit bits of BARG as a pointer
;	and the remaining 7 explicit bits as the argument to linear interpolation.


		MOVLW		HIGH (IBTBL256I)	; access table for F0
		MOVWF		TBLPTRH
		RLCF		BARGB1,W
		RLCF		BARGB0,W
		ADDLW		LOW (IBTBL256I)
		MOVWF		TBLPTRL
		BTFSC		_C
		INCF		TBLPTRH,F
		TABLRD		0,1,TEMPB0
		TLRD		1,TEMPB0
		TABLRD		0,0,TEMPB1
		TLRD		0,AARGB5

		MOVFP		AARGB5,WREG		; calculate difference
		SUBWF		TEMPB1,W
		MOVWF		AARGB5

		BCF		_C			; interpolate
		RLCF		BARGB2,W
		MULWF		AARGB5
		MOVPF		PRODH,TBLPTRH
		RLCF		BARGB1,W
		MULWF		AARGB5
		MOVPF		PRODL,WREG
		ADDWF		TBLPTRH,F
		BTFSC		_C
		INCF		PRODH,F

		CLRF		TEMPB2,F
		MOVFP		TBLPTRH,WREG
		SUBWF		TEMPB2,F
		MOVPF		PRODH,WREG
		SUBWFB		TEMPB1,F
		CLRF		WREG,F
		SUBWFB		TEMPB0,F		; F0

		endm

;----------------------------------------------------------------------------------------------

FPD32SEEDM	macro

;       Timing:     51 clks

;	PM: 49+129 = 178

;	generation of F0 by interpolating between consecutive 16 bit approximations
;	to the reciprocal of BARG, with the top 7 explicit bits of BARG as a pointer
;	and the remaining 16 explicit bits as the argument to linear interpolation.


		MOVLW		HIGH (IBTBL128I)	; access table for F0
		MOVWF		TBLPTRH
		MOVFP		BARGB0,WREG
		ANDLW		0x7F
		ADDLW		LOW (IBTBL128I)
		MOVWF		TBLPTRL
		BTFSC		_C
		INCF		TBLPTRH,F
		TABLRD		0,1,TEMPB0
		TLRD		1,TEMPB0
		TABLRD		0,0,TEMPB1
		TLRD		1,AARGB4
		TLRD		0,AARGB5

		MOVFP		AARGB5,WREG		; calculate difference
		SUBWF		TEMPB1,W
		MOVWF		AARGB5
		MOVFP		AARGB4,WREG
		SUBWFB		TEMPB0,W
		MOVWF		AARGB4

		MOVFP		AARGB5,WREG
		MOVWF		TEMPB2	

		MOVFP		AARGB5,WREG
		MULWF		BARGB2
		MOVPF		PRODH,TBLPTRH
		
		MOVFP		AARGB4,WREG
		MULWF		BARGB1
		MOVPF		PRODH,AARGB4
		MOVPF		PRODL,AARGB5

		MULWF		BARGB2
		MOVPF		PRODL,WREG
		ADDWF		TBLPTRH,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB5,F
		CLRF		WREG,F
		ADDWFC		AARGB4,F

		MOVFP		TEMPB2,WREG
		MULWF		BARGB1
		MOVPF		PRODL,WREG
		ADDWF		TBLPTRH,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB5,F
		CLRF		WREG,F
		ADDWFC		AARGB4,F

		CLRF		TEMPB2,F
		MOVFP		TBLPTRH,WREG
		SUBWF		TEMPB2,F
		MOVFP		AARGB5,WREG
		SUBWFB		TEMPB1,F
		MOVFP		AARGB4,WREG
		SUBWFB		TEMPB0,F		; F0

		endm

;----------------------------------------------------------------------------------------------

FPD32SEEDS	macro

;       Timing:     75 clks

;	PM: 73+65 = 138

;	generation of F0 by interpolating between consecutive 16 bit approximations
;	to the reciprocal of BARG, with the top 6 explicit bits of BARG as a pointer
;	and the remaining 17 explicit bits as the argument to linear interpolation.


		MOVLW		HIGH (IBTBL64I)	; access table for F0
		MOVWF		TBLPTRH
		RRCF		BARGB0,W
		ANDLW		0x3F
		ADDLW		LOW (IBTBL64I)
		MOVWF		TBLPTRL
		BTFSC		_C
		INCF		TBLPTRH,F
		TABLRD		0,1,TEMPB0
		TLRD		1,TEMPB0
		TABLRD		0,0,TEMPB1
		TLRD		1,AARGB4
		TLRD		0,AARGB5

		MOVFP		AARGB5,WREG		; calculate difference
		SUBWF		TEMPB1,W
		MOVWF		AARGB5
		MOVFP		AARGB4,WREG
		SUBWFB		TEMPB0,W
		MOVWF		AARGB4

		MOVFP		AARGB4,TBLPTRH	
		MOVFP		AARGB5,TBLPTRL	

		MOVFP		BARGB2,WREG
		MULWF		AARGB5
		MOVPF		PRODH,TMR0H
		MOVPF		PRODL,TMR0L
		
		MOVFP		BARGB1,WREG
		MULWF		AARGB4
		MOVPF		PRODH,AARGB5
		MOVPF		PRODL,TEMPB2

		MULWF		TBLPTRL
		MOVPF		PRODL,WREG
		ADDWF		TMR0H,F
		MOVPF		PRODH,WREG
		ADDWFC		TEMPB2,F
		CLRF		WREG,F
		ADDWFC		AARGB5,F

		MOVFP		BARGB2,WREG
		MULWF		TBLPTRH
		MOVPF		PRODL,WREG
		ADDWF		TMR0H,F
		MOVPF		PRODH,WREG
		ADDWFC		TEMPB2,F
		CLRF		WREG,F
		ADDWFC		AARGB5,F
	
		MOVFP		BARGB0,WREG
		ANDLW		0x01
		MULWF		TBLPTRL
		MOVPF		PRODL,WREG
		ADDWF		TEMPB2,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB5,F
		MOVFP		BARGB0,WREG
		ANDLW		0x01
		MULWF		TBLPTRH
		CLRF		AARGB4,W
		ADDWFC		AARGB4,F
		MOVPF		PRODL,WREG
		ADDWF		AARGB5,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB4,F

		BCF		_C
		RRCF		AARGB4,F
		RRCF		AARGB5,F
		RRCF		TEMPB2,F
		RRCF		TMR0H,F

		MOVFP		TEMPB2,TBLPTRH
		CLRF		TEMPB2,F
		MOVFP		TMR0H,WREG
		SUBWF		TEMPB2,F
		MOVFP		TBLPTRH,WREG
		SUBWFB		TEMPB1,F
		MOVFP		AARGB5,WREG
		SUBWFB		TEMPB0,F		; F0

		endm

;----------------------------------------------------------------------------------------------

PROG	CODE

		EXTERN		SETFUN32
		EXTERN		SETFDZ32
		EXTERN		RES032
		EXTERN		SETFOV32

FPD32
		GLOBAL		FPD32

		CLRF  		TEMPB3,W		; clear exponent modification
		CPFSGT		BEXP                    ; test for divide by zero
		GOTO  		SETFDZ32

		CPFSGT		AEXP
		GOTO  		RES032

D32BNE0
		MOVFP		AARGB0,WREG
		XORWF		BARGB0,W
		MOVPF		WREG,SIGN               ; save sign in SIGN

		BSF  		AARGB0,MSB              ; make argument MSB's explicit
		BSF  		BARGB0,MSB

;----------------------------------------------------------------------------------------------

		IF	DIVIDE == FAST

		FPD32SEEDF							; generation of F0

		ENDIF

;----------------------------------------------------------------------------------------------

		IF DIVIDE == MEDIUM

		FPD32SEEDM

		ENDIF

;----------------------------------------------------------------------------------------------

		IF DIVIDE == SLOW

		FPD32SEEDS

		ENDIF

;----------------------------------------------------------------------------------------------

		MOVFP		AARGB0,WREG
		MOVWF		AARGB5		; A0 = F0 * A	
		MOVFP		AARGB1,TBLPTRH	
		MOVFP		AARGB2,TBLPTRL	
		
		MOVFP		AARGB2,WREG
		MULWF		TEMPB2
		MOVPF		PRODH,AARGB4
		
		MOVFP		AARGB1,WREG
		MULWF		TEMPB1
		MOVPF		PRODH,AARGB2
		MOVPF		PRODL,AARGB3

		MULWF		TEMPB2
		MOVPF		PRODL,WREG
		ADDWF		AARGB4,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB3,F
		CLRF		WREG,F
		ADDWFC		AARGB2,F

		MOVFP		TBLPTRL,WREG
		MULWF		TEMPB1
		MOVPF		PRODL,WREG
		ADDWF		AARGB4,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB3,F
		CLRF		WREG,F
		ADDWFC		AARGB2,F
	
		MOVFP		AARGB0,WREG
		MULWF		TEMPB2
		MOVPF		PRODL,WREG
		ADDWF		AARGB3,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB2,F
		MOVFP		AARGB0,WREG
		MULWF		TEMPB1
		CLRF		AARGB1,W
		ADDWFC		AARGB1,F
		MOVPF		PRODL,WREG
		ADDWF		AARGB2,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB1,F

		MOVFP		TBLPTRL,WREG
		MULWF		TEMPB0
		MOVPF		PRODL,WREG
		ADDWF		AARGB3,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB2,F
		CLRF		AARGB0,W
		ADDWFC		AARGB1,F
		ADDWFC		AARGB0,F

		MOVFP		TBLPTRH,WREG
		MULWF		TEMPB0
		MOVPF		PRODL,WREG
		ADDWF		AARGB2,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB1,F
		CLRF		WREG,F
		ADDWFC		AARGB0,F

		MOVFP		AARGB5,WREG
		MULWF		TEMPB0
		MOVPF		PRODL,WREG
		ADDWF		AARGB1,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB0,F
	
		BTFSC		AARGB0,MSB
		GOTO		DAOK32
		RLCF		AARGB3,F
		RLCF		AARGB2,F
		RLCF		AARGB1,F
		RLCF		AARGB0,F
		DECF		TEMPB3,F

DAOK32
		MOVFP		BARGB0,WREG
		MOVWF		AARGB5		; B0 = F0 * B	
		MOVFP		BARGB1,TBLPTRH	
		MOVFP		BARGB2,TBLPTRL	
		
		MOVFP		BARGB2,WREG
		MULWF		TEMPB2
		MOVPF		PRODH,AARGB4
		
		MOVFP		BARGB1,WREG
		MULWF		TEMPB1
		MOVPF		PRODH,BARGB2
		MOVPF		PRODL,BARGB3

		MULWF		TEMPB2
		MOVPF		PRODL,WREG
		ADDWF		AARGB4,F
		MOVPF		PRODH,WREG
		ADDWFC		BARGB3,F
		CLRF		WREG,F
		ADDWFC		BARGB2,F

		MOVFP		TBLPTRL,WREG
		MULWF		TEMPB1
		MOVPF		PRODL,WREG
		ADDWF		AARGB4,F
		MOVPF		PRODH,WREG
		ADDWFC		BARGB3,F
		CLRF		WREG,F
		ADDWFC		BARGB2,F
	
		MOVFP		BARGB0,WREG
		MULWF		TEMPB2
		MOVPF		PRODL,WREG
		ADDWF		BARGB3,F
		MOVPF		PRODH,WREG
		ADDWFC		BARGB2,F
		MOVFP		BARGB0,WREG
		MULWF		TEMPB1
		CLRF		BARGB1,W
		ADDWFC		BARGB1,F
		MOVPF		PRODL,WREG
		ADDWF		BARGB2,F
		MOVPF		PRODH,WREG
		ADDWFC		BARGB1,F

		MOVFP		TBLPTRL,WREG
		MULWF		TEMPB0
		MOVPF		PRODL,WREG
		ADDWF		BARGB3,F
		MOVPF		PRODH,WREG
		ADDWFC		BARGB2,F
		CLRF		BARGB0,W
		ADDWFC		BARGB1,F
		ADDWFC		BARGB0,F

		MOVFP		TBLPTRH,WREG
		MULWF		TEMPB0
		MOVPF		PRODL,WREG
		ADDWF		BARGB2,F
		MOVPF		PRODH,WREG
		ADDWFC		BARGB1,F
		CLRF		WREG,F
		ADDWFC		BARGB0,F

		MOVFP		AARGB5,WREG
		MULWF		TEMPB0
		MOVPF		PRODL,WREG
		ADDWF		BARGB1,F
		MOVPF		PRODH,WREG
		ADDWFC		BARGB0,F
	
		BTFSS		BARGB0,MSB
		BTFSC		BARGB0,MSB-1
		GOTO		DBOK32
		RLCF		AARGB4,F
		RLCF		BARGB3,F
		RLCF		BARGB2,F
		RLCF		BARGB1,F
		RLCF		BARGB0,F
		INCF		TEMPB3,F

DBOK32
		COMF		BARGB3,F
		COMF		BARGB2,F
		COMF		BARGB1,F		; F1 = 2 - B0
		COMF		BARGB0,F
		INCF		BARGB3,F
		CLRF		WREG,F
		ADDWFC		BARGB2,F
		ADDWFC		BARGB1,F
		ADDWFC		BARGB0,F

		MOVFP		AARGB0,WREG
		MOVWF		TEMPB0		; A1 = F1 * A0
		MOVFP		AARGB1,WREG
		MOVWF		TEMPB1	
		MOVFP		AARGB2,WREG
		MOVWF		TEMPB2	
		MOVFP		AARGB3,TBLPTRL	
		
		MOVFP		AARGB2,WREG
		MULWF		BARGB2
		MOVPF		PRODH,AARGB4
	
		MOVFP		AARGB1,WREG
		MULWF		BARGB3
		MOVPF		PRODH,WREG
		ADDWF		AARGB4,F
		MOVFP		AARGB1,WREG
		MULWF		BARGB2
		CLRF		AARGB3,W
		ADDWFC		AARGB3,F
		MOVPF		PRODL,WREG
		ADDWF		AARGB4,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB3,F

		MOVFP		TBLPTRL,WREG
		MULWF		BARGB1
		MOVPF		PRODH,WREG
		ADDWF		AARGB4,F
		CLRF		AARGB2,W
		ADDWFC		AARGB3,F
		ADDWFC		AARGB2,F

		MOVFP		TEMPB2,WREG
		MULWF		BARGB1
		MOVPF		PRODL,WREG
		ADDWF		AARGB4,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB3,F
		CLRF		WREG,F
		ADDWFC		AARGB2,F

		MOVFP		TEMPB1,WREG
		MULWF		BARGB1
		MOVPF		PRODL,WREG
		ADDWF		AARGB3,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB2,F

		MOVFP		AARGB0,WREG
		MULWF		BARGB2
		MOVPF		PRODL,WREG
		ADDWF		AARGB3,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB2,F
		MOVFP		AARGB0,WREG
		MULWF		BARGB1
		CLRF		AARGB1,W
		ADDWFC		AARGB1,F
		MOVPF		PRODL,WREG
		ADDWF		AARGB2,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB1,F

		MOVFP		TEMPB0,WREG
		MULWF		BARGB3
		MOVPF		PRODL,WREG
		ADDWF		AARGB4,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB3,F
		CLRF		WREG,F
		ADDWFC		AARGB2,F
		ADDWFC		AARGB1,F

		MOVFP		TEMPB0,WREG
		MULWF		BARGB0
		MOVPF		PRODH,AARGB0
		MOVPF		PRODL,WREG
		ADDWF		AARGB1,F
		CLRF		WREG,F
		ADDWFC		AARGB0,F

		MOVFP		TBLPTRL,WREG
		MULWF		BARGB0
		MOVPF		PRODL,WREG
		ADDWF		AARGB4,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB3,F
		CLRF		WREG,F
		ADDWFC		AARGB2,F
		ADDWFC		AARGB1,F
		ADDWFC		AARGB0,F

		MOVFP		TEMPB2,WREG
		MULWF		BARGB0
		MOVPF		PRODL,WREG
		ADDWF		AARGB3,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB2,F
		CLRF		WREG,F
		ADDWFC		AARGB1,F
		ADDWFC		AARGB0,F

		MOVFP		TEMPB1,WREG
		MULWF		BARGB0
		MOVPF		PRODL,WREG
		ADDWF		AARGB2,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB1,F
		CLRF		WREG,F
		ADDWFC		AARGB0,F

		BTFSC		AARGB0,MSB
		GOTO		DEXP32
		RLCF		AARGB3,F
		RLCF		AARGB2,F
		RLCF		AARGB1,F
		RLCF		AARGB0,F
		DECF		TEMPB3,F

		BTFSC		AARGB0,MSB
		GOTO		DEXP32
		RLCF		AARGB3,F
		RLCF		AARGB2,F
		RLCF		AARGB1,F
		RLCF		AARGB0,F
		DECF		TEMPB3,F

DEXP32
		MOVFP		BEXP,WREG               ; compute AEXP - BEXP
		SUBWF		AEXP,F
		MOVLW		EXPBIAS+1		; add bias + 1 for scaling of F0
		BTFSS		_C
		GOTO 		ALTB32
        
AGEB32
		ADDWF		TEMPB3,W		; if AEXP > BEXP, test for overflow
		ADDWF		AEXP,F
		BTFSC		_C
		GOTO 		SETFOV32
		GOTO 		DROUND32

ALTB32
		ADDWF		TEMPB3,W		; if AEXP < BEXP, test for underflow
		ADDWF		AEXP,F
		BTFSS		_C
		GOTO 		SETFUN32

DROUND32
		BTFSC 		FPFLAGS,RND		; is rounding enabled?
		BTFSS 		AARGB3,MSB		; is NSB > 0x80?
		GOTO  		DIV32OK
		BSF	  		_C			; set carry for rounding
		MOVLW 		0x80
		CPFSGT		AARGB3			; if NSB = 0x80, select even
		RRCF  		AARGB2,W			; using lsb in carry
		CLRF  		WREG,F
		ADDWFC		AARGB2,F
		ADDWFC		AARGB1,F
		ADDWFC		AARGB0,F

		BTFSS 		_C                      ; test if rounding caused carryout
		GOTO  		DIV32OK
		RRCF  		AARGB0,F
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		INFSNZ		AEXP,F			; test for overflow
		GOTO  		SETFOV32

DIV32OK
		BTFSS		SIGN,MSB
		BCF  		AARGB0,MSB		; clear explicit MSB if positive

		RETLW		0

;----------------------------------------------------------------------------------------------

		IF DIVIDE == FAST

;	table for F0 by interpolating between consecutive 16 bit approximations
;	to the reciprocal of BARG, with the top 8 explicit bits of BARG as a pointer
;	and the remaining 7 explicit bits as the argument to linear interpolation.

IBTBL256I
		DATA	0xFFFF
		DATA	0xFF01
		DATA	0xFE04
		DATA	0xFD09
		DATA	0xFC10
		DATA	0xFB19
		DATA	0xFA23
		DATA	0xF930
		DATA	0xF83E
		DATA	0xF74E
		DATA	0xF660
		DATA	0xF574
		DATA	0xF48A
		DATA	0xF3A1
		DATA	0xF2BA
		DATA	0xF1D5
		DATA	0xF0F1
		DATA	0xF00F
		DATA	0xEF2F
		DATA	0xEE50
		DATA	0xED73
		DATA	0xEC98
		DATA	0xEBBE
		DATA	0xEAE5
		DATA	0xEA0F
		DATA	0xE939
		DATA	0xE866
		DATA	0xE793
		DATA	0xE6C3
		DATA	0xE5F3
		DATA	0xE526
		DATA	0xE459
		DATA	0xE38E
		DATA	0xE2C5
		DATA	0xE1FC
		DATA	0xE136
		DATA	0xE070
		DATA	0xDFAC
		DATA	0xDEE9
		DATA	0xDE28
		DATA	0xDD68
		DATA	0xDCA9
		DATA	0xDBEB
		DATA	0xDB2F
		DATA	0xDA74
		DATA	0xD9BA
		DATA	0xD902
		DATA	0xD84A
		DATA	0xD794
		DATA	0xD6DF
		DATA	0xD62C
		DATA	0xD579
		DATA	0xD4C7
		DATA	0xD417
		DATA	0xD368
		DATA	0xD2BA
		DATA	0xD20D
		DATA	0xD161
		DATA	0xD0B7
		DATA	0xD00D
		DATA	0xCF64
		DATA	0xCEBD
		DATA	0xCE17
		DATA	0xCD71
		DATA	0xCCCD
		DATA	0xCC29
		DATA	0xCB87
		DATA	0xCAE6
		DATA	0xCA46
		DATA	0xC9A6
		DATA	0xC908
		DATA	0xC86A
		DATA	0xC7CE
		DATA	0xC733
		DATA	0xC698
		DATA	0xC5FE
		DATA	0xC566
		DATA	0xC4CE
		DATA	0xC437
		DATA	0xC3A1
		DATA	0xC30C
		DATA	0xC278
		DATA	0xC1E5
		DATA	0xC152
		DATA	0xC0C1
		DATA	0xC030
		DATA	0xBFA0
		DATA	0xBF11
		DATA	0xBE83
		DATA	0xBDF6
		DATA	0xBD69
		DATA	0xBCDD
		DATA	0xBC52
		DATA	0xBBC8
		DATA	0xBB3F
		DATA	0xBAB6
		DATA	0xBA2F
		DATA	0xB9A8
		DATA	0xB921
		DATA	0xB89C
		DATA	0xB817
		DATA	0xB793
		DATA	0xB710
		DATA	0xB68D
		DATA	0xB60B
		DATA	0xB58A
		DATA	0xB50A
		DATA	0xB48A
		DATA	0xB40B
		DATA	0xB38D
		DATA	0xB30F
		DATA	0xB292
		DATA	0xB216
		DATA	0xB19B
		DATA	0xB120
		DATA	0xB0A6
		DATA	0xB02C
		DATA	0xAFB3
		DATA	0xAF3B
		DATA	0xAEC3
		DATA	0xAE4C
		DATA	0xADD6
		DATA	0xAD60
		DATA	0xACEB
		DATA	0xAC77
		DATA	0xAC03
		DATA	0xAB8F
		DATA	0xAB1D
		DATA	0xAAAB
		DATA	0xAA39
		DATA	0xA9C8
		DATA	0xA958
		DATA	0xA8E8
		DATA	0xA879
		DATA	0xA80B
		DATA	0xA79C
		DATA	0xA72F
		DATA	0xA6C2
		DATA	0xA656
		DATA	0xA5EA
		DATA	0xA57F
		DATA	0xA514
		DATA	0xA4AA
		DATA	0xA440
		DATA	0xA3D7
		DATA	0xA36E
		DATA	0xA306
		DATA	0xA29F
		DATA	0xA238
		DATA	0xA1D1
		DATA	0xA16B
		DATA	0xA106
		DATA	0xA0A1
		DATA	0xA03C
		DATA	0x9FD8
		DATA	0x9F74
		DATA	0x9F11
		DATA	0x9EAF
		DATA	0x9E4D
		DATA	0x9DEB
		DATA	0x9D8A
		DATA	0x9D29
		DATA	0x9CC9
		DATA	0x9C69
		DATA	0x9C0A
		DATA	0x9BAB
		DATA	0x9B4C
		DATA	0x9AEE
		DATA	0x9A91
		DATA	0x9A34
		DATA	0x99D7
		DATA	0x997B
		DATA	0x991F
		DATA	0x98C4
		DATA	0x9869
		DATA	0x980E
		DATA	0x97B4
		DATA	0x975A
		DATA	0x9701
		DATA	0x96A8
		DATA	0x9650
		DATA	0x95F8
		DATA	0x95A0
		DATA	0x9549
		DATA	0x94F2
		DATA	0x949C
		DATA	0x9446
		DATA	0x93F0
		DATA	0x939B
		DATA	0x9346
		DATA	0x92F1
		DATA	0x929D
		DATA	0x9249
		DATA	0x91F6
		DATA	0x91A3
		DATA	0x9150
		DATA	0x90FE
		DATA	0x90AC
		DATA	0x905A
		DATA	0x9009
		DATA	0x8FB8
		DATA	0x8F68
		DATA	0x8F17
		DATA	0x8EC8
		DATA	0x8E78
		DATA	0x8E29
		DATA	0x8DDA
		DATA	0x8D8C
		DATA	0x8D3E
		DATA	0x8CF0
		DATA	0x8CA3
		DATA	0x8C56
		DATA	0x8C09
		DATA	0x8BBC
		DATA	0x8B70
		DATA	0x8B24
		DATA	0x8AD9
		DATA	0x8A8E
		DATA	0x8A43
		DATA	0x89F8
		DATA	0x89AE
		DATA	0x8964
		DATA	0x891B
		DATA	0x88D2
		DATA	0x8889
		DATA	0x8840
		DATA	0x87F8
		DATA	0x87AF
		DATA	0x8768
		DATA	0x8720
		DATA	0x86D9
		DATA	0x8692
		DATA	0x864C
		DATA	0x8605
		DATA	0x85BF
		DATA	0x8579
		DATA	0x8534
		DATA	0x84EF
		DATA	0x84AA
		DATA	0x8465
		DATA	0x8421
		DATA	0x83DD
		DATA	0x8399
		DATA	0x8356
		DATA	0x8312
		DATA	0x82CF
		DATA	0x828D
		DATA	0x824A
		DATA	0x8208
		DATA	0x81C6
		DATA	0x8185
		DATA	0x8143
		DATA	0x8102
		DATA	0x80C1
		DATA	0x8081
		DATA	0x8040
		DATA	0x8001

		ENDIF

;----------------------------------------------------------------------------------------------

		IF DIVIDE == MEDIUM

;	generation of F0 by interpolating between consecutive 16 bit approximations
;	to the reciprocal of BARG, with the top 7 explicit bits of BARG as a pointer
;	and the remaining 16 explicit bits as the argument to linear interpolation.

IBTBL128I
		DATA	0xFFFF
		DATA	0xFE04
		DATA	0xFC10
		DATA	0xFA23
		DATA	0xF83E
		DATA	0xF660
		DATA	0xF48A
		DATA	0xF2BA
		DATA	0xF0F1
		DATA	0xEF2F
		DATA	0xED73
		DATA	0xEBBE
		DATA	0xEA0F
		DATA	0xE866
		DATA	0xE6C3
		DATA	0xE526
		DATA	0xE38E
		DATA	0xE1FC
		DATA	0xE070
		DATA	0xDEE9
		DATA	0xDD68
		DATA	0xDBEB
		DATA	0xDA74
		DATA	0xD902
		DATA	0xD794
		DATA	0xD62C
		DATA	0xD4C7
		DATA	0xD368
		DATA	0xD20D
		DATA	0xD0B7
		DATA	0xCF64
		DATA	0xCE17
		DATA	0xCCCD
		DATA	0xCB87
		DATA	0xCA46
		DATA	0xC908
		DATA	0xC7CE
		DATA	0xC698
		DATA	0xC566
		DATA	0xC437
		DATA	0xC30C
		DATA	0xC1E5
		DATA	0xC0C1
		DATA	0xBFA0
		DATA	0xBE83
		DATA	0xBD69
		DATA	0xBC52
		DATA	0xBB3F
		DATA	0xBA2F
		DATA	0xB921
		DATA	0xB817
		DATA	0xB710
		DATA	0xB60B
		DATA	0xB50A
		DATA	0xB40B
		DATA	0xB30F
		DATA	0xB216
		DATA	0xB120
		DATA	0xB02C
		DATA	0xAF3B
		DATA	0xAE4C
		DATA	0xAD60
		DATA	0xAC77
		DATA	0xAB8F
		DATA	0xAAAB
		DATA	0xA9C8
		DATA	0xA8E8
		DATA	0xA80B
		DATA	0xA72F
		DATA	0xA656
		DATA	0xA57F
		DATA	0xA4AA
		DATA	0xA3D7
		DATA	0xA306
		DATA	0xA238
		DATA	0xA16B
		DATA	0xA0A1
		DATA	0x9FD8
		DATA	0x9F11
		DATA	0x9E4D
		DATA	0x9D8A
		DATA	0x9CC9
		DATA	0x9C0A
		DATA	0x9B4C
		DATA	0x9A91
		DATA	0x99D7
		DATA	0x991F
		DATA	0x9869
		DATA	0x97B4
		DATA	0x9701
		DATA	0x9650
		DATA	0x95A0
		DATA	0x94F2
		DATA	0x9446
		DATA	0x939B
		DATA	0x92F1
		DATA	0x9249
		DATA	0x91A3
		DATA	0x90FE
		DATA	0x905A
		DATA	0x8FB8
		DATA	0x8F17
		DATA	0x8E78
		DATA	0x8DDA
		DATA	0x8D3E
		DATA	0x8CA3
		DATA	0x8C09
		DATA	0x8B70
		DATA	0x8AD9
		DATA	0x8A43
		DATA	0x89AE
		DATA	0x891B
		DATA	0x8889
		DATA	0x87F8
		DATA	0x8768
		DATA	0x86D9
		DATA	0x864C
		DATA	0x85BF
		DATA	0x8534
		DATA	0x84AA
		DATA	0x8421
		DATA	0x8399
		DATA	0x8312
		DATA	0x828D
		DATA	0x8208
		DATA	0x8185
		DATA	0x8102
		DATA	0x8081
		DATA	0x8001

		ENDIF

;----------------------------------------------------------------------------------------------

		IF	DIVIDE == SLOW

;	table for interpolating between consecutive 16 bit approximations
;	to the reciprocal of BARG, with the top 6 explicit bits of BARG as a pointer
;	and the remaining 17 explicit bits as the argument to linear interpolation.

IBTBL64I
		DATA	0xFFFF
		DATA	0xFC10
		DATA	0xF83E
		DATA	0xF48A
		DATA	0xF0F1
		DATA	0xED73
		DATA	0xEA0F
		DATA	0xE6C3
		DATA	0xE38E
		DATA	0xE070
		DATA	0xDD68
		DATA	0xDA74
		DATA	0xD794
		DATA	0xD4C7
		DATA	0xD20D
		DATA	0xCF64
		DATA	0xCCCD
		DATA	0xCA46
		DATA	0xC7CE
		DATA	0xC566
		DATA	0xC30C
		DATA	0xC0C1
		DATA	0xBE83
		DATA	0xBC52
		DATA	0xBA2F
		DATA	0xB817
		DATA	0xB60B
		DATA	0xB40B
		DATA	0xB216
		DATA	0xB02C
		DATA	0xAE4C
		DATA	0xAC77
		DATA	0xAAAB
		DATA	0xA8E8
		DATA	0xA72F
		DATA	0xA57F
		DATA	0xA3D7
		DATA	0xA238
		DATA	0xA0A1
		DATA	0x9F11
		DATA	0x9D8A
		DATA	0x9C0A
		DATA	0x9A91
		DATA	0x991F
		DATA	0x97B4
		DATA	0x9650
		DATA	0x94F2
		DATA	0x939B
		DATA	0x9249
		DATA	0x90FE
		DATA	0x8FB8
		DATA	0x8E78
		DATA	0x8D3E
		DATA	0x8C09
		DATA	0x8AD9
		DATA	0x89AE
		DATA	0x8889
		DATA	0x8768
		DATA	0x864C
		DATA	0x8534
		DATA	0x8421
		DATA	0x8312
		DATA	0x8208
		DATA	0x8102
		DATA	0x8001

		ENDIF

		END
