;	RCS Header $Id: FPM32.ASM,v 1.2 2000/02/22 22:15:27 krt Exp $

;	$Revision: 1.2 $

	include <P17C44.INC>		; General 17CXXX definitions

	include <CMATH17.INC>		; Math library definitions

;**********************************************************************************************
;**********************************************************************************************

;       Floating Point Multiply

;       Input:  32 bit floating point number in AEXP, AARGB0, AARGB1, AARGB2
;               32 bit floating point number in BEXP, BARGB0, BARGB1, BARGB2

;       Use:    CALL    FPM32

;       Output: 32 bit floating point product in AEXP, AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  AARG * BARG

;	Testing on [MINNUM,MAXNUM] from 100000 trials:

;			min	max		mean
;	Timing:	14	2024	1834.7	clks

;			min		max		mean	rms
;	Error:	-0xA2	0x9A	2.20	29.18	nsb

;       PM: 123						DM: 15


;----------------------------------------------------------------------------------------------

PROG	CODE

		EXTERN		RES032
		EXTERN		SETFOV32
		EXTERN		SETFUN32

FPM32
		GLOBAL		FPM32

		CLRF  		AARGB3,W                ; test for zero arguments
	    CPFSEQ		BEXP
	    CPFSGT		AEXP
	    GOTO  		RES032

M32BNE0
		MOVFP		AARGB0,WREG
		XORWF 		BARGB0,W
		MOVPF 		WREG,SIGN               ; save sign in SIGN

		MOVFP 		BEXP,WREG
		ADDWF 		AEXP,F
		MOVLW 		EXPBIAS-1
		BTFSS 		_C
		GOTO  		MTUN32

		SUBWF 		AEXP,F			; remove bias and overflow test
		BTFSC 		_C
		GOTO  		SETFOV32
		GOTO  		MOK32

MTUN32
		SUBWF 		AEXP,F			; remove bias and underflow test
		BTFSS 		_C
		GOTO  		SETFUN32

MOK32
		BSF         AARGB0,MSB              ; make argument MSB's explicit
		BSF			BARGB0,MSB

		MOVFP		AARGB0,WREG
		MOVWF		TEMPB0		; multiply mantissas	
		MOVFP		AARGB1,WREG
		MOVWF		TEMPB1	
		MOVFP		AARGB2,WREG
		MOVWF		TEMPB2	

		MOVFP		AARGB2,WREG
		MULWF		BARGB2
		MOVPF		PRODH,AARGB4

		MOVFP		AARGB1,WREG
		MULWF		BARGB1
		MOVPF		PRODH,AARGB2
		MOVPF		PRODL,AARGB3

		MULWF		BARGB2
		MOVPF		PRODL,WREG
		ADDWF		AARGB4,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB3,F
		CLRF		WREG,F
		ADDWFC		AARGB2,F

		MOVFP		TEMPB2,WREG
		MULWF		BARGB1
		MOVPF		PRODL,WREG
		ADDWF		AARGB4,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB3,F
		CLRF		WREG,F
		ADDWFC		AARGB2,F

		MOVFP		AARGB0,WREG
		MULWF		BARGB2
		MOVPF		PRODL,WREG
		ADDWF		AARGB3,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB2,F
		MOVFP		AARGB0,WREG
		MULWF		BARGB1
		CLRF		AARGB1,W
		ADDWFC		AARGB1,F
		MOVPF		PRODL,WREG
		ADDWF		AARGB2,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB1,F

		MOVFP		TEMPB2,WREG
		MULWF		BARGB0
		MOVPF		PRODL,WREG
		ADDWF		AARGB3,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB2,F
		CLRF		AARGB0,W
		ADDWFC		AARGB1,F
		ADDWFC		AARGB0,F

		MOVFP		TEMPB1,WREG
		MULWF		BARGB0
		MOVPF		PRODL,WREG
		ADDWF		AARGB2,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB1,F
		CLRF		WREG,F
		ADDWFC		AARGB0,F

		MOVFP		TEMPB0,WREG
		MULWF		BARGB0
		MOVPF		PRODL,WREG
		ADDWF		AARGB1,F
		MOVPF		PRODH,WREG
		ADDWFC		AARGB0,F

		BTFSC		AARGB0,MSB		; check for postnormalization
		GOTO		MROUND32
		RLCF		AARGB3,F
		RLCF		AARGB2,F
		RLCF		AARGB1,F
		RLCF		AARGB0,F
		DECF		AEXP,F
		BTFSC		_Z
		GOTO		SETFUN32

MROUND32
		BTFSC       FPFLAGS,RND		; is rounding enabled?
		BTFSS 		AARGB3,MSB		; is NSB > 0x80?
		GOTO  		MUL32OK
		BSF	  		_C			; set carry for rounding
		MOVLW 		0x80
		CPFSGT		AARGB3			; if NSB = 0x80, select even
		RRCF  		AARGB2,W		; using lsb in carry
		CLRF  		WREG,F
		ADDWFC		AARGB2,F
		ADDWFC		AARGB1,F
		ADDWFC		AARGB0,F

		BTFSS 		_C                      ; has rounding caused carryout?
		GOTO  		MUL32OK
		RRCF  		AARGB0,F                ; if so, right shift
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		INFSNZ		AEXP,F			; test for floating point overflow
		GOTO  		SETFOV32

MUL32OK
		BTFSS		SIGN,MSB
		BCF  		AARGB0,MSB		; clear explicit MSB if positive

		RETLW		0  

		END
