;	RCS Header $Id: INT3208.ASM,v 1.2 2000/02/22 22:15:28 krt Exp $

;	$Revision: 1.2 $

	include <P17C44.INC>		; General 17CXXX definitions

	include <CMATH17.INC>		; Math library definitions

;**********************************************************************************************
;**********************************************************************************************

;       32 bit float to 8 bit integer conversion

;       Input:  32 bit floating point number in AEXP, AARGB0, AARGB1, AARGB2

;       Use:    CALL    INT3208

;       Output: 8 bit signed integer right justified in AARGB0

;       Result: AARG  <--  INT( AARG )

;	Testing on [-127.0,127.0] from 100000 trials:

;		min	max	mean
;	Timing:	27	52	35.8	clks

;		min	max	mean	rms
;	Error:	0x00	0x00	0.00	0.00	nsb

;       PM: 84						DM: 6

;----------------------------------------------------------------------------------------------

PROG	CODE

		EXTERN		SETIOV3208

INT3208
		GLOBAL		INT3208

		CLRF  		WREG,F
		CPFSGT		AEXP			; test for zero argument
		RETLW 		0x00
		MOVFP 		AARGB0,WREG
		MOVWF 		SIGN			; save sign in SIGN
		BSF   		AARGB0,MSB		; make MSB explicit

		MOVLW 		EXPBIAS+D'7'		; remove bias from EXP
		SUBWF 		AEXP,W

		BTFSS 		WREG,MSB		; if >= 7, integer overflow
		GOTO  		SETIOV3208		; will occur
		NEGW  		AEXP,F

		MOVLW 		0x07			; do byte shift if EXP >= 8
		CPFSGT		AEXP
		GOTO  		SNIB3208
		SUBWF 		AEXP,F			; EXP = EXP - 7
		MOVFP 		AARGB0,WREG
		MOVWF 		AARGB1			; save for rounding
		CLRF  		AARGB0,F
		DCFSNZ		AEXP,F			; EXP = EXP - 1
		GOTO  		SHIFT3208OK             ; shift completed if EXP = 0

SNIB3208A
		MOVLW 		3                       ; do nibble shift if EXP >= 4
		CPFSGT		AEXP
		GOTO  		SHIFT3208A
		SUBWF 		AEXP,F			; EXP = EXP - 3
		SWAPF 		AARGB1,W		; save for rounding
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB1

		DCFSNZ		AEXP,F			; EXP = EXP - 1
		GOTO  		SHIFT3208OK             ; shift completed if EXP = 0

SHIFT3208A
		BCF   		_C                      ; at most 3 right shifts are required
		RRCF  		AARGB1,F                ; right shift by EXP
		DCFSNZ		AEXP,F
		GOTO  		SHIFT3208OK             ; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB1,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT3208OK             ; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB1,F
		GOTO  		SHIFT3208OK

SNIB3208
		MOVLW 		3                       ; do nibble shift if EXP >= 4
		CPFSGT		AEXP
		GOTO  		SHIFT3208
		SUBWF 		AEXP,F			; EXP = EXP - 3
		SWAPF 		AARGB1,W		; save for rounding
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB1

		SWAPF 		AARGB0,W
		ANDLW 		0xF0
		ADDWF 		AARGB1,F

		SWAPF 		AARGB0,W
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB0
		DCFSNZ		AEXP,F			; EXP = EXP - 1
		GOTO  		SHIFT3208OK             ; shift completed if EXP = 0

SHIFT3208
		BCF   		_C                      ; at most 3 right shifts are required
		RRCF  		AARGB0,F                ; right shift by EXP
		RRCF  		AARGB1,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT3208OK             ; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB0,F 
		RRCF  		AARGB1,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT3208OK             ; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB0,F 
		RRCF  		AARGB1,F

SHIFT3208OK
		BTFSC 		FPFLAGS,RND		; is rounding enabled?
		BTFSS 		AARGB1,MSB		; is NSB > 0x80?
		GOTO  		INT3208OK
		BSF	  		_C			; set carry for rounding
		MOVLW 		0x80
		CPFSGT		AARGB1			; if NSB = 0x80, select even
		RRCF  		AARGB0,W		; using lsb in carry
		CLRF  		WREG,F
		ADDWFC		AARGB0,F
		BTFSC 		AARGB0,MSB
		GOTO  		SETIOV3208

INT3208OK
		BTFSS		SIGN,MSB                ; if sign bit set, negate               
		RETLW		0x00
		COMF 		AARGB0,F
		INCF 		AARGB0,F
		RETLW		0x00

		END
