;	RCS Header $Id: INT3216.ASM,v 1.2 2000/02/22 22:15:28 krt Exp $

;	$Revision: 1.2 $

	include <P17C44.INC>		; General 17CXXX definitions

	include <CMATH17.INC>		; Math library definitions

;**********************************************************************************************
;**********************************************************************************************

;       32 bit float to 16 bit integer conversion

;       Input:  32 bit floating point number in AEXP, AARGB0, AARGB1, AARGB2

;       Use:    CALL    INT3216

;       Output: 16 bit signed integer right justified in AARGB0, AARGB1

;       Result: AARG  <--  INT( AARG )

;	Testing on [-32767.0,32767.0] from 100000 trials:

;			min	max	mean
;	Timing:	29	64	39.4	clks

;			min	max	mean	rms
;	Error:	0x00	0x00	0.00	0.00	nsb

;       PM: 137                                         DM: 6

;----------------------------------------------------------------------------------------------

PROG	CODE

		EXTERN		SETIOV3216

INT3216
		GLOBAL		INT3216

		CLRF  		WREG,F
		CPFSGT		AEXP				; test for zero argument
		RETLW 		0x00
		MOVFP 		AARGB0,WREG
		MOVWF 		SIGN			; save sign in SIGN
		BSF   		AARGB0,MSB		; make MSB explicit

		MOVLW 		EXPBIAS+D'15'		; remove bias from EXP
		SUBWF 		AEXP,W

		BTFSS 		WREG,MSB		; if >= 15, integer overflow
		GOTO  		SETIOV3216		; will occur
		NEGW  		AEXP,F

		MOVLW 		0x07			; do byte shift if EXP >= 8
		CPFSGT		AEXP
		GOTO  		SNIB3216
		SUBWF 		AEXP,F			; EXP = EXP - 7
		MOVFP 		AARGB1,WREG
		MOVWF 		AARGB2			; save for rounding
		MOVFP 		AARGB0,WREG
		MOVWF 		AARGB1
		CLRF  		AARGB0,F
		DCFSNZ		AEXP,F			; EXP = EXP - 1
		GOTO  		SHIFT3216OK             ; shift completed if EXP = 0

		MOVLW 		0x07
		CPFSGT		AEXP                     ; do another byte shift if EXP >= 8
		GOTO  		SNIB3216A
		SUBWF 		AEXP,F			; EXP = EXP - 7
		MOVFP 		AARGB1,WREG
		MOVWF 		AARGB2			; save for rounding
		CLRF  		AARGB1,F
		DCFSNZ		AEXP,F			; EXP = EXP - 1
		GOTO  		SHIFT3216OK             ; shift completed if EXP = 0

SNIB3216B
		MOVLW 		3                       ; do nibble shift if EXP >= 4
		CPFSGT		AEXP
		GOTO  		SHIFT3216B
		SUBWF 		AEXP,F			; EXP = EXP - 3

		SWAPF 		AARGB2,W		; save for rounding
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB2
		DCFSNZ		AEXP,F			; EXP = EXP - 1
		GOTO  		SHIFT3216OK             ; shift completed if EXP = 0

SHIFT3216B
		BCF   		_C                      ; at most 3 right shifts are required
		RRCF  		AARGB2,F                ; right shift by EXP
		DCFSNZ		AEXP,F
		GOTO  		SHIFT3216OK             ; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB2,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT3216OK             ; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB2,F
		GOTO  		SHIFT3216OK

SNIB3216A
		MOVLW 		3                       ; do nibble shift if EXP >= 4
		CPFSGT		AEXP
		GOTO  		SHIFT3216A
		SUBWF 		AEXP,F			; EXP = EXP - 3
		SWAPF 		AARGB2,W		; save for rounding
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB2

		SWAPF 		AARGB1,W
		ANDLW 		0xF0
		ADDWF 		AARGB2,F

		SWAPF 		AARGB1,W
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB1
		DCFSNZ		AEXP,F			; EXP = EXP - 1
		GOTO  		SHIFT3216OK             ; shift completed if EXP = 0

SHIFT3216A
		BCF   		_C                      ; at most 3 right shifts are required
		RRCF  		AARGB1,F                ; right shift by EXP
		RRCF  		AARGB2,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT3216OK             ; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT3216OK             ; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		GOTO  		SHIFT3216OK

SNIB3216
		MOVLW 		3                       ; do nibble shift if EXP >= 4
		CPFSGT		AEXP
		GOTO  		SHIFT3216
		SUBWF 		AEXP,F			; EXP = EXP - 3
		SWAPF 		AARGB2,W		; save for rounding
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB2

		SWAPF 		AARGB1,W
		ANDLW 		0xF0
		ADDWF 		AARGB2,F

		SWAPF 		AARGB1,W
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB1

		SWAPF 		AARGB0,W
		ANDLW 		0xF0
		ADDWF 		AARGB1,F

		SWAPF 		AARGB0,W
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB0
		DCFSNZ		AEXP,F			; EXP = EXP - 1
		GOTO  		SHIFT3216OK             ; shift completed if EXP = 0

SHIFT3216
		BCF   		_C                      ; at most 3 right shifts are required
		RRCF  		AARGB0,F                ; right shift by EXP
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT3216OK             ; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB0,F 
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT3216OK             ; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB0,F 
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F

SHIFT3216OK
		BTFSC 		FPFLAGS,RND		; is rounding enabled?
		BTFSS 		AARGB2,MSB		; is NSB > 0x80?
		GOTO  		INT3216OK
		BSF	  		_C			; set carry for rounding
		MOVLW 		0x80
		CPFSGT		AARGB2			; if NSB = 0x80, select even
		RRCF  		AARGB1,W		; using lsb in carry
		CLRF  		WREG,F
		ADDWFC		AARGB1,F
		ADDWFC		AARGB0,F
		BTFSC 		AARGB0,MSB
		GOTO  		SETIOV3216

INT3216OK
		BTFSS 		SIGN,MSB                ; if sign bit set, negate               
		RETLW 		0x00
		COMF  		AARGB1,F
		COMF  		AARGB0,F
		INCF  		AARGB1,F
		CLRF  		WREG,F
		ADDWFC		AARGB0,F
		RETLW 		0x00

		END
