;	RCS Header $Id: INT3232.ASM,v 1.3 2000/03/15 19:12:43 krt Exp $

;	$Revision: 1.3 $

	include <P17C44.INC>		; General 17CXXX definitions

	include <CMATH17.INC>		; Math library definitions

;**********************************************************************************************
;**********************************************************************************************

;       Float to integer conversion

;       Input:  32 bit floating point number in AEXP, AARGB0, AARGB1, AARGB2

;       Use:    CALL    INT3232

;       Output: 32 bit 2's complement integer right justified in AARGB0, AARGB1, AARGB2,
;               AARGB3

;       Result: AARG  <--  INT( AARG )

;	Testing on [MINNUM,MAXNUM] from 10000 trials:

;			min	max	mean

;	Timing:	35	83	47.4	clks

;       PM: 249                                         DM: 9

;----------------------------------------------------------------------------------------------

PROG	CODE

		EXTERN		SETIOV3232

INT3232
		GLOBAL		INT3232

		CLRF  		AARGB3,W
		CPFSGT		AEXP			 		; test for zero argument
		RETLW 		0x00
		MOVFP 		AARGB0,WREG
		MOVWF		SIGN	 		; save sign in SIGN
		BSF   		AARGB0,MSB	 		; make MSB explicit

		CLRF  		AARGB4,F

		MOVLW 		EXPBIAS+D'31'		; remove bias from EXP
		SUBWF 		AEXP,W

		BTFSS 		WREG,MSB	 		; if >= 31, integer overflow
		GOTO  		SETIOV3232	 		; will occur
		NEGW  		AEXP,F

		MOVLW 		7            		; do byte shift if EXP >= 8
		CPFSGT		AEXP
		GOTO  		SNIB3232
		SUBWF 		AEXP,F		 		; EXP = EXP - 7
		MOVFP 		AARGB3,WREG
		MOVWF		AARGB4		; save for rounding
		MOVFP 		AARGB2,WREG
		MOVWF		AARGB3
		MOVFP 		AARGB1,WREG
		MOVWF		AARGB2
		MOVFP 		AARGB0,WREG
		MOVWF		AARGB1
		CLRF  		AARGB0,F
		DCFSNZ		AEXP,F		 		; EXP = EXP - 1
		GOTO  		SHIFT3232OK  		; shift completed if EXP = 0

		MOVLW 		7            		; do byte shift if EXP >= 8
		CPFSGT		AEXP          		; do another byte shift if EXP >= 8
		GOTO  		SNIB3232A
		SUBWF 		AEXP,F		 		; EXP = EXP - 7
		MOVFP 		AARGB3,WREG
		MOVWF		AARGB4		; save for rounding
		MOVFP 		AARGB2,WREG
		MOVWF		AARGB3
		MOVFP 		AARGB1,WREG
		MOVWF		AARGB2
		CLRF  		AARGB1,F
		DCFSNZ		AEXP,F		 		; EXP = EXP - 1
		GOTO  		SHIFT3232OK  		; shift completed if EXP = 0

		MOVLW 		7            		; do byte shift if EXP >= 8
		CPFSGT		AEXP          		; do another byte shift if EXP >= 8
		GOTO  		SNIB3232B
		SUBWF 		AEXP,F		 		; EXP = EXP - 7
		MOVFP 		AARGB3,WREG
		MOVWF		AARGB4		; save for rounding
		MOVFP 		AARGB2,WREG
		MOVWF		AARGB3
		CLRF  		AARGB2,F
		DCFSNZ		AEXP,F		 		; EXP = EXP - 1
		GOTO  		SHIFT3232OK  		; shift completed if EXP = 0

		MOVLW 		7            		; do byte shift if EXP >= 8
		CPFSGT		AEXP          		; do another byte shift if EXP >= 8
		GOTO  		SNIB3232C
		SUBWF 		AEXP,F		 		; EXP = EXP - 7
		MOVFP 		AARGB3,WREG
		MOVWF		AARGB4		; save for rounding
		CLRF  		AARGB3,F
		DCFSNZ		AEXP,F		 		; EXP = EXP - 1
		GOTO  		SHIFT3232OK  		; shift completed if EXP = 0

SNIB3232D
		MOVLW 		3          			; do nibble shift if EXP >= 4
		CPFSGT		AEXP
		GOTO  		SHIFT3232D
		SWAPF 		AARGB4,W
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB4
		GOTO  		SHIFT3232OK			; shift completed if EXP = 0

SHIFT3232D
		BCF   		_C                      ; at most 3 right shifts are required
		RRCF  		AARGB4,F                ; right shift by EXP
		DCFSNZ		AEXP,F
		GOTO  		SHIFT3232OK             ; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB4,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT3232OK             ; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB4,F
		GOTO  		SHIFT3232OK

SNIB3232C
		MOVLW 		3                       ; do nibble shift if EXP >= 4
		CPFSGT		AEXP
		GOTO  		SHIFT3232C
		SUBWF 		AEXP,F			; EXP = EXP - 3
		SWAPF 		AARGB3,W
		MOVPF 		WREG,AARGB4		; save for rounding
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB3
		DCFSNZ		AEXP,F			; EXP = EXP - 1
		GOTO  		SHIFT3232OK             ; shift completed if EXP = 0

SHIFT3232C
		BCF   		_C                      ; at most 3 right shifts are required
		RRCF  		AARGB3,F                ; right shift by EXP
		RRCF  		AARGB4,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT3232OK             ; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB3,F
		RRCF  		AARGB4,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT3232OK             ; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB3,F
		RRCF  		AARGB4,F
		GOTO  		SHIFT3232OK

SNIB3232B
		MOVLW 		3                       ; do nibble shift if EXP >= 4
		CPFSGT		AEXP
		GOTO  		SHIFT3232B
		SUBWF 		AEXP,F			; EXP = EXP - 3
		SWAPF 		AARGB3,W
		MOVPF 		WREG,AARGB4		; save for rounding
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB3

		SWAPF 		AARGB2,W
		ANDLW 		0xF0
		ADDWF 		AARGB3,F

		SWAPF 		AARGB2,W
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB2
		DCFSNZ		AEXP,F			; EXP = EXP - 1
		GOTO  		SHIFT3232OK             ; shift completed if EXP = 0

SHIFT3232B
		BCF   		_C                      ; at most 3 right shifts are required
		RRCF  		AARGB2,F                ; right shift by EXP
		RRCF  		AARGB3,F
		RRCF  		AARGB4,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT3232OK             ; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		RRCF  		AARGB4,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT3232OK             ; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		RRCF  		AARGB4,F
		GOTO  		SHIFT3232OK

SNIB3232A
		MOVLW 		3                       ; do nibble shift if EXP >= 4
		CPFSGT		AEXP
		GOTO  		SHIFT3232A
		SUBWF 		AEXP,F			; EXP = EXP - 3
		SWAPF 		AARGB3,W
		MOVPF 		WREG,AARGB4		; save for rounding
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB3

		SWAPF 		AARGB2,W
		ANDLW 		0xF0
		ADDWF 		AARGB3,F

		SWAPF 		AARGB2,W
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB2

		SWAPF 		AARGB1,W
		ANDLW 		0xF0
		ADDWF 		AARGB2,F

		SWAPF 		AARGB1,W
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB1
		DCFSNZ		AEXP,F			; EXP = EXP - 1
		GOTO  		SHIFT3232OK             ; shift completed if EXP = 0

SHIFT3232A
		BCF   		_C                      ; at most 3 right shifts are required
		RRCF  		AARGB1,F                ; right shift by EXP
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		RRCF  		AARGB4,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT3232OK             ; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		RRCF  		AARGB4,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT3232OK             ; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		RRCF  		AARGB4,F
		GOTO  		SHIFT3232OK

SNIB3232
		MOVLW 		3                       ; do nibble shift if EXP >= 4
		CPFSGT		AEXP
		GOTO  		SHIFT3232
		SUBWF 		AEXP,F			; EXP = EXP - 3
		SWAPF 		AARGB3,W
		MOVPF 		WREG,AARGB4		; save for rounding
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB3

		SWAPF 		AARGB2,W
		ANDLW 		0xF0
		ADDWF 		AARGB3,F

		SWAPF 		AARGB2,W
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB2

		SWAPF 		AARGB1,W
		ANDLW 		0xF0
		ADDWF 		AARGB2,F

		SWAPF 		AARGB1,W
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB1

		SWAPF 		AARGB0,W
		ANDLW 		0xF0
		ADDWF 		AARGB1,F

		SWAPF 		AARGB0,W
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB0
		DCFSNZ		AEXP,F			; EXP = EXP - 1
		GOTO  		SHIFT3232OK             ; shift completed if EXP = 0

SHIFT3232
		BCF   		_C                      ; at most 3 right shifts are required
		RRCF  		AARGB0,F                ; right shift by EXP
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		RRCF  		AARGB4,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT3232OK             ; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB0,F 
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		RRCF  		AARGB4,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT3232OK             ; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB0,F 
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		RRCF  		AARGB4,F

SHIFT3232OK
		BTFSC 		FPFLAGS,RND		; is rounding enabled?
		BTFSS 		AARGB4,MSB		; is NSB > 0x80?
		GOTO  		INT3232OK
		BSF	  		_C			; set carry for rounding
		MOVLW 		0x80
		CPFSGT		AARGB4			; if NSB = 0x80, select even
		RRCF  		AARGB3,W		; using lsb in carry
		CLRF  		WREG,F
		ADDWFC		AARGB3,F
		ADDWFC		AARGB2,F
		ADDWFC		AARGB1,F
		ADDWFC		AARGB0,F
		BTFSC 		AARGB0,MSB
		GOTO  		SETIOV3232

INT3232OK
		BTFSS 		SIGN,MSB                ; if sign bit set, negate               
		RETLW 		0
		COMF  		AARGB3,F
		COMF  		AARGB2,F
		COMF  		AARGB1,F
		COMF  		AARGB0,F
		INCF  		AARGB3,F
		CLRF  		WREG,F
		ADDWFC		AARGB2,F
		ADDWFC		AARGB1,F
		ADDWFC		AARGB0,F
		RETLW 		0

		END
