;   RCS Header $Id: NORM32.ASM,v 1.1 1999/01/08 19:56:36 GrosbaJ Exp $

;   $Revision: 1.1 $

;**********************************************************************************************
;**********************************************************************************************

;       Normalization routine

;       Input:  40 bit unnormalized floating point number in AEXP, AARGB0, AARGB1,
;               AARGB2, AARGB3 with sign in SIGN,MSB.

;       Use:    CALL    NRM4032

;       Output: 32 bit normalized floating point number in AEXP, AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  NORMALIZE( AARG )


;       Max Timing:     27+22+8+4 = 61 clks             RND = 0
;                       27+22+24+4 = 77 clks            RND = 1, SAT = 0
;                       27+22+23+11 = 83 clks           RND = 1, SAT = 1

;       Min Timing:     4+8+8+6+6 = 32 clks             AARG = 0
;                       3+5+4+8+6 = 26 clks

;       PM: 178                                         DM: 8

;----------------------------------------------------------------------------------------------

        #include    "cmath17.inc"

        EXTERN      SETFUN32, FIXSIGN32, SETFOV32, RES032

WREG        EQU     H'0A'

NORM32_CODE  CODE

NRM4032         
        GLOBAL      NRM4032

        PAGESEL     NRM4032
        BANKSEL     AARGB0

        CLRF            TEMP,W          ; clear exponent decrement
                CPFSGT          AARGB0          ; test if highbyte=0
                GOTO            NRM4032A

TNIB4032        MOVLW           0xF0                    ; test if highnibble=0
                ANDWF           AARGB0,W
                TSTFSZ          WREG
                GOTO            NORM4032
                SWAPF           AARGB0,F                ; if so, shift 4 bits
                SWAPF           AARGB1,W
                ANDLW           0x0F
                ADDWF           AARGB0,F

                SWAPF           AARGB1,W
                ANDLW           0xF0
                MOVPF           WREG,AARGB1
                SWAPF           AARGB2,W
                ANDLW           0x0F
                ADDWF           AARGB1,F

                SWAPF           AARGB2,W
                ANDLW           0xF0
                MOVPF           WREG,AARGB2
                SWAPF           AARGB3,W
                ANDLW           0x0F
                ADDWF           AARGB2,F

                SWAPF           AARGB3,W
                ANDLW           0xF0
                MOVPF           WREG,AARGB3

                BSF             TEMP,2                  ; increase decrement by 4

NORM4032        BCF             _C                      ; clear carry bit

                BTFSC           AARGB0,MSB      ; if MSB=1, normalization done
                GOTO            TNORMUN4032
                RLCF            AARGB3,F                ; otherwise, shift left and 
                RLCF            AARGB2,F                ; increment decrement
                RLCF            AARGB1,F
                RLCF            AARGB0,F
                INCF            TEMP,F
                BTFSC           AARGB0,MSB
                GOTO            TNORMUN4032
                RLCF            AARGB3,F
                RLCF            AARGB2,F
                RLCF            AARGB1,F
                RLCF            AARGB0,F
                INCF            TEMP,F
                BTFSC           AARGB0,MSB      ; since highnibble != 0, at most
                GOTO            TNORMUN4032             ; 3 left shifts are required
                RLCF            AARGB3,F
                RLCF            AARGB2,F
                RLCF            AARGB1,F
                RLCF            AARGB0,F
                INCF            TEMP,F

TNORMUN4032     MOVFP           TEMP,WREG               ; if EXP <= decrement in TEMP,
                CPFSGT          EXP                     ; floating point underflow has
                GOTO            SETFUN32                ; occured
                SUBWF           EXP,F           ; otherwise, compute EXP

NRMRND4032
        GLOBAL      NRMRND4032

        BTFSC           FPFLAGS,RND     ; is rounding enabled?
                BTFSS           AARGB3,MSB      ; is NSB > 0x80?
                GOTO            FIXSIGN32
        BSF     _C          ; set carry for rounding
        MOVLW       0x80
        CPFSGT      AARGB3          ; if NSB = 0x80, select even
        RRCF        AARGB2,W        ; using lsb in carry
                CLRF            WREG,F
                ADDWFC          AARGB2,F
                ADDWFC          AARGB1,F
                ADDWFC          AARGB0,F

                BTFSS           _C                      ; has rounding caused carryout?
                GOTO            FIXSIGN32
                RRCF            AARGB0,F                ; if so, right shift
                RRCF            AARGB1,F
                RRCF            AARGB2,F
                INFSNZ          EXP,F           ; test for floating point overflow
                GOTO            SETFOV32
                GOTO            FIXSIGN32

NRM4032A        MOVFP           AARGB1,WREG
        MOVWF       AARGB0          ; shift 8 bits by move
                MOVFP           AARGB2,WREG
        MOVWF       AARGB1
                MOVFP           AARGB3,WREG
        MOVWF       AARGB2
                CLRF            AARGB3,W
                BSF             TEMP,3                  ; increase decrement by 8       
                CPFSGT          AARGB0          ; test if highbyte=0
                GOTO            NRM4032B

TNIB4032A       MOVLW           0xF0                    ; test if highnibble=0
                ANDWF           AARGB0,W
                TSTFSZ          WREG
                GOTO            NORM4032A
                SWAPF           AARGB0,F                ; if so, shift 4 bits
                SWAPF           AARGB1,W
                ANDLW           0x0F
                ADDWF           AARGB0,F

                SWAPF           AARGB1,W
                ANDLW           0xF0
                MOVPF           WREG,AARGB1
                SWAPF           AARGB2,W
                ANDLW           0x0F
                ADDWF           AARGB1,F

                SWAPF           AARGB2,W
                ANDLW           0xF0
                MOVPF           WREG,AARGB2

                BSF             TEMP,2                  ; increase decrement by 4

NORM4032A       BCF             _C          ; clear carry bit

                BTFSC           AARGB0,MSB      ; if MSB=1, normalization done
                GOTO            TNORMUN4032
                RLCF            AARGB2,F                ; otherwise, shift left and 
                RLCF            AARGB1,F                ; increment decrement
                RLCF            AARGB0,F
                INCF            TEMP,F
                BTFSC           AARGB0,MSB
                GOTO            TNORMUN4032
                RLCF            AARGB2,F
                RLCF            AARGB1,F
                RLCF            AARGB0,F
                INCF            TEMP,F
                BTFSC           AARGB0,MSB      ; since highnibble != 0, at most
                GOTO            TNORMUN4032             ; 3 left shifts are required
                RLCF            AARGB2,F
                RLCF            AARGB1,F
                RLCF            AARGB0,F
                INCF            TEMP,F
                GOTO            TNORMUN4032

NRM4032B        MOVFP           AARGB1,WREG
        MOVWF       AARGB0          ; shift 8 bits by move
                MOVFP           AARGB2,WREG
        MOVWF       AARGB1
                CLRF            AARGB2,W
                BCF             TEMP,3                  ; increase decrement by 8
                BSF             TEMP,4  
                CPFSGT          AARGB0          ; if highbyte=0, result=0
                GOTO            NRM4032C

TNIB4032B       MOVLW           0xF0                    ; test if highnibble=0
                ANDWF           AARGB0,W
                TSTFSZ          WREG
                GOTO            NORM4032B

                SWAPF           AARGB0,F                ; if so, shift 4 bits
                SWAPF           AARGB1,W
                ANDLW           0x0F
                ADDWF           AARGB0,F

                SWAPF           AARGB1,W
                ANDLW           0xF0
                MOVPF           WREG,AARGB1

                BSF             TEMP,2                  ; increase decrement by 4

NORM4032B       BCF             _C                      ; clear carry bit

                BTFSC           AARGB0,MSB      ; if MSB=1, normalization done
                GOTO            TNORMUN4032
                RLCF            AARGB1,F                ; otherwise, shift left and 
                RLCF            AARGB0,F
                INCF            TEMP,F          ; increment decrement
                BTFSC           AARGB0,MSB
                GOTO            TNORMUN4032
                RLCF            AARGB1,F
                RLCF            AARGB0,F
                INCF            TEMP,F
                BTFSC           AARGB0,MSB      ; since highnibble != 0, at most
                GOTO            TNORMUN4032             ; 3 left shifts are required
                RLCF            AARGB1,F
                RLCF            AARGB0,F
                INCF            TEMP,F
                GOTO            TNORMUN4032

NRM4032C        MOVFP           AARGB1,WREG
        MOVWF       AARGB0          ; shift 8 bits by move
                CLRF            AARGB1,W
                BSF             TEMP,3                  ; increase decrement by 8
                CPFSGT          AARGB0          ; if highbyte=0, result=0
                GOTO            RES032

TNIB4032C       MOVLW           0xF0                    ; test if highnibble=0
                ANDWF           AARGB0,W
                TSTFSZ          WREG
                GOTO            NORM4032C
                SWAPF           AARGB0,F                ; if so, shift 4 bits

                BSF             TEMP,2                  ; increase decrement by 4

NORM4032C       BCF             _C                      ; clear carry bit

                BTFSC           AARGB0,MSB      ; if MSB=1, normalization done
                GOTO            TNORMUN4032
                RLCF            AARGB0,F                ; otherwise, shift left and 
                INCF            TEMP,F          ; increment decrement
                BTFSC           AARGB0,MSB
                GOTO            TNORMUN4032
                RLCF            AARGB0,F
                INCF            TEMP,F
                BTFSC           AARGB0,MSB      ; since highnibble != 0, at most
                GOTO            TNORMUN4032             ; 3 left shifts are required
                RLCF            AARGB0,F
                INCF            TEMP,F
                GOTO            TNORMUN4032


        END
