;	RCS Header $Id: fpas32.asm,v 1.1 2000/03/15 19:21:47 krt Exp $

;	$Revision: 1.1 $

	include <P17C44.INC>		; General 17CXXX definitions

	include <CMATH17.INC>		; Math library definitions

;**********************************************************************************************
;**********************************************************************************************

;       Floating Point Subtract

;       Input:  32 bit floating point number in AEXP, AARGB0, AARGB1, AARGB2
;               32 bit floating point number in BEXP, BARGB0, BARGB1, BARGB2

;       Use:    CALL FPS32

;       Output: 32 bit floating point difference in AEXP, AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  AARG - BARG

;	Testing on [MINNUM,MAXNUM] from 100000 trials:

;			min		max		mean
;	Timing:	54		123		80.2	clks

;			min		max		mean	rms
;	Error:	0x00	0x01	0.0008	0.03	nsb

;       PM: 425                                 DM: 14

;----------------------------------------------------------------------------------------------

PROG	CODE

		EXTERN		NRM4032
		EXTERN		SETFOV32
		EXTERN		NRMRND4032

FPS32
		GLOBAL		FPS32

		BTG			BARGB0,MSB             ; toggle sign bit for subtraction

;**********************************************************************************************

;       Floating Point Add

;       Input:  32 bit floating point number in AEXP, AARGB0, AARGB1, AARGB2
;               32 bit floating point number in BEXP, BARGB0, BARGB1, BARGB2

;       Use:    CALL FPA32

;       Output: 32 bit floating point sum in AEXP, AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  AARG - BARG

;	Testing on [MINNUM,MAXNUM] from 100000 trials:

;			min		max		mean
;	Timing:	53		122		79.2	clks

;			min		max		mean	rms
;	Error:	0x00	0x01	0.0008	0.03	nsb

;       PM: 424                                 DM: 14

;----------------------------------------------------------------------------------------------

FPA32
		GLOBAL		FPA32

		MOVFP 		AARGB0,WREG             ; exclusive or of signs in TEMPB0
		XORWF 		BARGB0,W
		MOVPF 		WREG,TEMPB0

		CLRF  		AARGB3,F

		MOVFP 		AEXP,WREG               ; use AARG if AEXP >= BEXP
		CPFSGT		BEXP
		GOTO  		USEA32

USEB32
		MOVFP 		BARGB0,WREG             ; use BARG if AEXP < BEXP
		MOVPF 		WREG,SIGN               ; save sign in SIGN
		BSF   		BARGB0,MSB              ; make MSB's explicit
		BSF   		AARGB0,MSB

		MOVFP 		AEXP,WREG               ; compute shift count in BEXP
		MOVPF 		WREG,TEMPB1
		MOVFP 		BEXP,WREG
		MOVPF 		WREG,AEXP

		CLRF  		WREG,F
		CPFSGT		TEMPB1			; return BARG if AARG = 0
		GOTO  		BRETURN32
		MOVFP 		TEMPB1,WREG
		SUBWF 		BEXP,F
		BTFSC 		_Z
		GOTO  		BLIGNED32

		MOVLW 		7
		CPFSGT		BEXP                    ; do byte shift if BEXP >= 8 
		GOTO  		BNIB32

		SUBWF 		BEXP,F			; BEXP = BEXP - 7
		MOVFP 		AARGB2,WREG
		MOVWF		AARGB3		; keep for postnormalization
		MOVFP 		AARGB1,WREG
		MOVWF		AARGB2
		MOVFP 		AARGB0,WREG
		MOVWF		AARGB1
		CLRF  		AARGB0,F
		DCFSNZ		BEXP,F			; BEXP = BEXP - 1
		GOTO  		BLIGNED32

		MOVLW 		7            		; do byte shift if EXP >= 8
		CPFSGT		BEXP                    ; do another byte shift if BEXP >= 8 
		GOTO  		BNIB32A
		SUBWF 		BEXP,F			; BEXP = BEXP - 7
		MOVFP 		AARGB2,WREG
		MOVWF		AARGB3		; keep for postnormalization
		MOVFP 		AARGB1,WREG
		MOVWF		AARGB2
		CLRF  		AARGB1,F
		DCFSNZ		BEXP,F			; BEXP = BEXP - 1
		GOTO  		BLIGNED32

		MOVLW 		7            		; do byte shift if EXP >= 8
		CPFSGT		BEXP                    ; do another byte shift if BEXP >= 8 
		GOTO  		BNIB32B
		SUBWF 		BEXP,F			; BEXP = BEXP - 7
		MOVFP 		AARGB2,WREG
		MOVWF		AARGB3		; keep for postnormalization
		CLRF  		AARGB2,F
		DCFSNZ		BEXP,F			; BEXP = BEXP - 1
		GOTO  		BLIGNED32

		MOVLW 		7            		; do byte shift if EXP >= 8
		CPFSGT		BEXP                    ; if BEXP still >= 8, then
		GOTO  		BNIB32C                 ; AARG = 0 relative to BARG

BRETURN32
		MOVFP		SIGN,WREG
		MOVWF		AARGB0             ; return BARG
		MOVFP		BARGB1,WREG
		MOVWF		AARGB1
		MOVFP		BARGB2,WREG
		MOVWF		AARGB2
		CLRF 		AARGB3,F
		RETLW		0x00

BNIB32C
		MOVLW 		3                       ; do nibbleshift if BEXP >= 4
		CPFSGT		BEXP
		GOTO  		BLOOP32C
		SUBWF 		BEXP,F			; BEXP = BEXP -3
		SWAPF 		AARGB3,W
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB3
		DCFSNZ		BEXP,F			; BEXP = BEXP - 1
		GOTO  		BLIGNED32		; aligned if BEXP = 0

BLOOP32C
		BCF   		_C                      ; right shift by BEXP
		RRCF  		AARGB3,F
		DCFSNZ		BEXP,F
		GOTO  		BLIGNED32               ; aligned if BEXP = 0
		BCF   		_C
		RRCF  		AARGB3,F
		DCFSNZ		BEXP,F
		GOTO  		BLIGNED32               ; aligned if BEXP = 0
		BCF   		_C                      ; at most 3 right shifts are
		RRCF  		AARGB3,F                ; possible
		GOTO  		BLIGNED32

BNIB32B
		MOVLW 		3                       ; do nibbleshift if BEXP >= 4
		CPFSGT		BEXP
		GOTO  		BLOOP32B
		SUBWF 		BEXP,F			; BEXP = BEXP -3
		SWAPF 		AARGB3,W
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB3
		SWAPF 		AARGB2,W
		ANDLW 		0xF0
		ADDWF 		AARGB3,F
		SWAPF 		AARGB2,W
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB2
		DCFSNZ		BEXP,F			; BEXP = BEXP - 1
		GOTO  		BLIGNED32		; aligned if BEXP = 0

BLOOP32B
		BCF   		_C                      ; right shift by BEXP
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		DCFSNZ		BEXP,F
		GOTO  		BLIGNED32               ; aligned if BEXP = 0
		BCF   		_C
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		DCFSNZ		BEXP,F
		GOTO  		BLIGNED32               ; aligned if BEXP = 0
		BCF   		_C                      ; at most 3 right shifts are
		RRCF  		AARGB2,F                ; possible
		RRCF  		AARGB3,F
		GOTO  		BLIGNED32

BNIB32A
		MOVLW 		3                       ; do nibbleshift if BEXP >= 4
		CPFSGT		BEXP
		GOTO  		BLOOP32A
		SUBWF 		BEXP,F			; BEXP = BEXP -3
		SWAPF 		AARGB3,W
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB3
		SWAPF 		AARGB2,W
		ANDLW 		0xF0
		ADDWF 		AARGB3,F
		SWAPF 		AARGB2,W
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB2
		SWAPF 		AARGB1,W
		ANDLW 		0xF0
		ADDWF 		AARGB2,F
		SWAPF 		AARGB1,W
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB1
		DCFSNZ		BEXP,F			; BEXP = BEXP - 1
		GOTO  		BLIGNED32               ; aligned if BEXP = 0

BLOOP32A
		BCF   		_C                      ; right shift by BEXP
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		DCFSNZ		BEXP,F
		GOTO  		BLIGNED32               ; aligned if BEXP = 0
		BCF   		_C
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		DCFSNZ		BEXP,F
		GOTO  		BLIGNED32		; aligned if BEXP = 0
		BCF   		_C                      ; at most 3 right shifts are
		RRCF  		AARGB1,F		; possible
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		GOTO  		BLIGNED32

BNIB32
		MOVLW 		3                       ; do nibbleshift if BEXP >= 4
		CPFSGT		BEXP
		GOTO  		BLOOP32
		SUBWF 		BEXP,F			; BEXP = BEXP -3
		SWAPF 		AARGB3,W
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB3
		SWAPF 		AARGB2,W
		ANDLW 		0xF0
		ADDWF 		AARGB3,F
		SWAPF 		AARGB2,W
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB2
		SWAPF 		AARGB1,W
		ANDLW 		0xF0
		ADDWF 		AARGB2,F
		SWAPF 		AARGB1,W
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB1
		SWAPF 		AARGB0,W
		ANDLW 		0xF0
		ADDWF 		AARGB1,F
		SWAPF 		AARGB0,W
		ANDLW 		0x0F
		MOVPF 		WREG,AARGB0
		DCFSNZ		BEXP,F			; BEXP = BEXP - 1
		GOTO  		BLIGNED32               ; aligned if BEXP = 0

BLOOP32
		BCF   		_C                      ; right shift by BEXP
		RRCF  		AARGB0,F
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		DCFSNZ		BEXP,F
		GOTO  		BLIGNED32               ; aligned if BEXP = 0
		BCF   		_C
		RRCF  		AARGB0,F
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		DCFSNZ		BEXP,F
		GOTO  		BLIGNED32               ; aligned if BEXP = 0
		BCF   		_C                      ; at most 3 right shifts are
		RRCF  		AARGB0,F                ; possible
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F

BLIGNED32
		CLRF  		BARGB3,W
		BTFSS 		TEMPB0,MSB              ; negate if signs opposite
		GOTO  		AOK32
		COMF  		AARGB3,F
		COMF  		AARGB2,F
		COMF  		AARGB1,F
		COMF  		AARGB0,F
		INCF  		AARGB3,F
		ADDWFC		AARGB2,F
		ADDWFC		AARGB1,F
		ADDWFC		AARGB0,F
		GOTO  		AOK32

USEA32
		TSTFSZ		BEXP                    ; return AARG if BARG = 0
		GOTO  		BNE032
		RETLW 		0x00

BNE032
		CLRF  		BARGB3,F
		MOVFP 		AARGB0,WREG
		MOVWF		SIGN             ; save sign in SIGN
		BSF   		AARGB0,MSB              ; make MSB's explicit
		BSF   		BARGB0,MSB

		MOVFP 		BEXP,WREG               ; compute shift count in BEXP
		SUBWF 		AEXP,W
		MOVPF 		WREG,BEXP
		BTFSC 		_Z
		GOTO  		ALIGNED32

		MOVLW 		7
		CPFSGT		BEXP                    ; do byte shift if BEXP >= 8 
		GOTO  		ANIB32
		SUBWF 		BEXP,F			; BEXP = BEXP - 7
		MOVFP 		BARGB2,WREG		; keep for postnormalization
		MOVPF 		WREG,BARGB3
		MOVFP 		BARGB1,WREG
		MOVPF 		WREG,BARGB2
		MOVFP 		BARGB0,WREG
		MOVPF 		WREG,BARGB1
		CLRF  		BARGB0,F
		DCFSNZ		BEXP,F			; BEXP = BEXP - 1
		GOTO  		ALIGNED32

		MOVLW 		7
		CPFSGT		BEXP                    ; do another byte shift if BEXP >= 8 
		GOTO  		ANIB32A
		SUBWF 		BEXP,F			; BEXP = BEXP - 7
		MOVFP 		BARGB2,WREG
		MOVPF 		WREG,BARGB3
		MOVFP 		BARGB1,WREG
		MOVPF 		WREG,BARGB2
		CLRF  		BARGB1,F
		DCFSNZ		BEXP,F			; BEXP = BEXP - 1
		GOTO  		ALIGNED32

		MOVLW 		7
		CPFSGT		BEXP                    ; do another byte shift if BEXP >= 8 
		GOTO  		ANIB32B
		SUBWF 		BEXP,F			; BEXP = BEXP - 7
		MOVFP 		BARGB2,WREG
		MOVPF 		WREG,BARGB3
		CLRF  		BARGB2,F
		DCFSNZ		BEXP,F			; BEXP = BEXP - 1
		GOTO  		ALIGNED32

		MOVLW 		7
		CPFSGT		BEXP                    ; if BEXP still >= 8, then
		GOTO  		ANIB32C                 ; BARG = 0 relative to AARG

		MOVFP 		SIGN,WREG
		MOVWF		AARGB0             ; return AARG
		RETLW 		0x00

ANIB32C
		MOVLW 		3                       ; do nibbleshift if BEXP >= 4
		CPFSGT		BEXP
		GOTO  		ALOOP32C
		SUBWF 		BEXP,F			; BEXP = BEXP -3
		SWAPF 		BARGB3,W
		ANDLW 		0x0F
		MOVPF 		WREG,BARGB3
		DCFSNZ		BEXP,F			; BEXP = BEXP - 1
		GOTO  		ALIGNED32               ; aligned if BEXP = 0

ALOOP32C
		BCF   		_C                      ; right shift by BEXP
		RRCF  		BARGB3,F
		DCFSNZ		BEXP,F
		GOTO  		ALIGNED32               ; aligned if BEXP = 0
		BCF   		_C
		RRCF  		BARGB3,F
		DCFSNZ		BEXP,F
		GOTO  		ALIGNED32               ; aligned if BEXP = 0
		BCF   		_C                      ; at most 3 right shifts are
		RRCF  		BARGB3,F                ; possible
		GOTO  		ALIGNED32

ANIB32B
		MOVLW 		3                       ; do nibbleshift if BEXP >= 4
		CPFSGT		BEXP
		GOTO  		ALOOP32B
		SUBWF 		BEXP,F			; BEXP = BEXP -3
		SWAPF 		BARGB3,W
		ANDLW 		0x0F
		MOVPF 		WREG,BARGB3
		SWAPF 		BARGB2,W
		ANDLW 		0xF0
		ADDWF 		BARGB3,F
		SWAPF 		BARGB2,W
		ANDLW 		0x0F
		MOVPF 		WREG,BARGB2
		DCFSNZ		BEXP,F			; BEXP = BEXP - 1
		GOTO  		ALIGNED32               ; aligned if BEXP = 0

ALOOP32B
		BCF   		_C                      ; right shift by BEXP
		RRCF  		BARGB2,F
		RRCF  		BARGB3,F
		DCFSNZ		BEXP,F
		GOTO  		ALIGNED32               ; aligned if BEXP = 0
		BCF   		_C
		RRCF  		BARGB2,F
		RRCF  		BARGB3,F
		DCFSNZ		BEXP,F
		GOTO  		ALIGNED32               ; aligned if BEXP = 0
		BCF   		_C                      ; at most 3 right shifts are
		RRCF  		BARGB2,F                ; possible
		RRCF  		BARGB3,F
		GOTO  		ALIGNED32

ANIB32A
		MOVLW 		3                       ; do nibbleshift if BEXP >= 4
		CPFSGT		BEXP
		GOTO  		ALOOP32A
		SUBWF 		BEXP,F			; BEXP = BEXP -3
		SWAPF 		BARGB3,W
		ANDLW 		0x0F
		MOVPF 		WREG,BARGB3
		SWAPF 		BARGB2,W
		ANDLW 		0xF0
		ADDWF 		BARGB3,F
		SWAPF 		BARGB2,W
		ANDLW 		0x0F
		MOVPF 		WREG,BARGB2
		SWAPF 		BARGB1,W
		ANDLW 		0xF0
		ADDWF 		BARGB2,F
		SWAPF 		BARGB1,W
		ANDLW 		0x0F
		MOVPF 		WREG,BARGB1
		DCFSNZ		BEXP,F			; BEXP = BEXP - 1
		GOTO  		ALIGNED32               ; aligned if BEXP = 0

ALOOP32A
		BCF   		_C                      ; right shift by BEXP
		RRCF  		BARGB1,F
		RRCF  		BARGB2,F
		RRCF  		BARGB3,F
		DCFSNZ		BEXP,F
		GOTO  		ALIGNED32               ; aligned if BEXP = 0
		BCF   		_C
		RRCF  		BARGB1,F
		RRCF  		BARGB2,F
		RRCF  		BARGB3,F
		DCFSNZ		BEXP,F
		GOTO  		ALIGNED32               ; aligned if BEXP = 0
		BCF   		_C                      ; at most 3 right shifts are
		RRCF  		BARGB1,F                ; possible
		RRCF  		BARGB2,F
		RRCF  		BARGB3,F
		GOTO  		ALIGNED32

ANIB32
		MOVLW 		3                       ; do nibbleshift if BEXP >= 4
		CPFSGT		BEXP
		GOTO  		ALOOP32
		SUBWF 		BEXP,F			; BEXP = BEXP -3
		SWAPF 		BARGB3,W
		ANDLW 		0x0F
		MOVPF 		WREG,BARGB3
		SWAPF 		BARGB2,W
		ANDLW 		0xF0
		ADDWF 		BARGB3,F
		SWAPF 		BARGB2,W
		ANDLW 		0x0F
		MOVPF 		WREG,BARGB2
		SWAPF 		BARGB1,W
		ANDLW 		0xF0
		ADDWF 		BARGB2,F
		SWAPF 		BARGB1,W
		ANDLW 		0x0F
		MOVPF 		WREG,BARGB1
		SWAPF 		BARGB0,W
		ANDLW 		0xF0
		ADDWF 		BARGB1,F
		SWAPF 		BARGB0,W
		ANDLW 		0x0F
		MOVPF 		WREG,BARGB0
		DCFSNZ		BEXP,F			; BEXP = BEXP - 1
		GOTO  		ALIGNED32               ; aligned if BEXP = 0

ALOOP32
		BCF   		_C                      ; right shift by BEXP
		RRCF  		BARGB0,F
		RRCF  		BARGB1,F
		RRCF  		BARGB2,F
		RRCF  		BARGB3,F
		DCFSNZ		BEXP,F
		GOTO  		ALIGNED32                ; aligned if BEXP = 0
		BCF   		_C
		RRCF  		BARGB0,F
		RRCF  		BARGB1,F
		RRCF  		BARGB2,F
		RRCF  		BARGB3,F
		DCFSNZ		BEXP,F
		GOTO  		ALIGNED32		; aligned if BEXP = 0
		BCF   		_C                      ; at most 3 right shifts are
		RRCF  		BARGB0,F		; possible
		RRCF  		BARGB1,F
		RRCF  		BARGB2,F
		RRCF  		BARGB3,F

ALIGNED32
		CLRF  		AARGB3,W
		BTFSS 		TEMPB0,MSB              ; negate if signs opposite
		GOTO  		AOK32
		COMF  		BARGB3,F
		COMF  		BARGB2,F
		COMF  		BARGB1,F
		COMF  		BARGB0,F
		INCF  		BARGB3,F
		ADDWFC		BARGB2,F
		ADDWFC		BARGB1,F
		ADDWFC		BARGB0,F

AOK32
		MOVFP 		BARGB3,WREG
		ADDWF 		AARGB3,F
		MOVFP 		BARGB2,WREG             ; add
		ADDWFC		AARGB2,F
		MOVFP 		BARGB1,WREG
		ADDWFC		AARGB1,F
		MOVFP 		BARGB0,WREG
		ADDWFC		AARGB0,F

		BTFSC 		TEMPB0,MSB
		GOTO  		ACOMP32
		BTFSS 		_C
		GOTO  		NRMRND4032

		RRCF  		AARGB0,F                ; shift right and increment EXP
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		INCFSZ		AEXP,F
		GOTO  		NRMRND4032
		GOTO  		SETFOV32                ; set floating point overflow flag

ACOMP32
		BTFSC 		_C
		GOTO  		NRM4032                 ; normalize and fix sign
		CLRF  		WREG,F
		COMF  		AARGB3,F                ; negate, toggle sign bit and
		COMF  		AARGB2,F                ; then normalize
		COMF  		AARGB1,F
		COMF  		AARGB0,F
		INCF  		AARGB3,F
		ADDWFC		AARGB2,F
		ADDWFC		AARGB1,F
		ADDWFC		AARGB0,F
		BTG   		SIGN,MSB
		GOTO  		NRM4032

		END
