;	RCS Header $Id: fxd2416u.asm,v 1.1 2000/03/14 20:34:58 krt Exp $

;	$Revision: 1.1 $

	include <P17C44.INC>		; General 17CXXX definitions

	include <CMATH17.INC>		; Math library definitions

;**********************************************************************************************
;**********************************************************************************************

;       24/16 Bit Unsigned Fixed Point Divide 24/16 -> 24.16

;       Input:  24 bit unsigned fixed point dividend in AARGB0, AARGB1, AARGB2
;               16 bit unsigned fixed point divisor in BARGB0, BARGB1

;       Use:    CALL    FXD2416U

;       Output: 24 bit unsigned fixed point quotient in AARGB0, AARGB1, AARGB2
;               16 bit unsigned fixed point remainder in REMB0, REMB1

;       Result: AARG, REM  <--  AARG / BARG

;	Testing from 100000 trials:

;		min	max	mean
;	Timing:	30	276	224.61	clks

;       PM: 135               DM: 15

;----------------------------------------------------------------------------------------------

PROG	CODE

		EXTERN		FXD2408U
		EXTERN		FXD1616U
		EXTERN		FXD1608U
		EXTERN		DGETNRMD
		EXTERN		FXD1608U2
		EXTERN		IBXTBL256

FXD2416U
		GLOBAL		FXD2416U

		CLRF		REMB0,F
		CLRF		REMB1,F
		CLRF		WREG,F
		TSTFSZ		BARGB0
		GOTO		D2416BGT1
		MOVFP		BARGB1,WREG
		MOVWF		BARGB0
		CALL		FXD2408U
		MOVFP		REMB0,WREG
		MOVWF		REMB1
		CLRF		REMB0,F
		RETLW		0x00

D2416BGT1
		CPFSEQ		AARGB0
		GOTO		D2416AGTB
		MOVFP		AARGB1,WREG
		MOVWF		AARGB0
		MOVFP		AARGB2,WREG
		MOVWF		AARGB1
		CALL		FXD1616U
		
		MOVFP		AARGB1,WREG
		MOVWF		AARGB2
		MOVFP		AARGB0,WREG
		MOVWF		AARGB1
		CLRF		AARGB0,F
		RETLW		0x00
D2416AGTB
		MOVFP		AARGB2,WREG
		MOVWF		AARGB3
		MOVFP		AARGB1,WREG
		MOVWF		AARGB2
		MOVFP		AARGB0,WREG
		MOVWF		AARGB1
		CLRF		AARGB0,F

		MOVFP		AARGB0,WREG
		MOVWF		TEMPB0
		MOVFP		AARGB1,WREG
		MOVWF		TEMPB1
		MOVFP		AARGB2,WREG
		MOVWF		TEMPB2
		MOVFP		AARGB3,WREG
		MOVWF		TEMPB3

		MOVLW		0x02		; set loop count
		MOVWF		AEXP

		MOVLW		0x01
		MOVWF		ZARGB0

		BTFSC		BARGB0,MSB
		GOTO		D2416UNRMOK

		CALL		DGETNRMD	; get normalization factor
		MOVWF		ZARGB0

		MULWF		BARGB1
		MOVFP		BARGB0,WREG
		MOVPF		PRODL,BARGB1
		MOVPF		PRODH,BARGB0
		MULWF		ZARGB0
		MOVFP		PRODL,WREG
		ADDWF		BARGB0,F

		MOVFP		ZARGB0,WREG
		MULWF		AARGB3
		MOVPF		PRODL,TEMPB3
		MOVPF		PRODH,TEMPB2
		MULWF		AARGB1
		MOVPF		PRODL,TEMPB1
		MOVPF		PRODH,TEMPB0
		MULWF		AARGB2
		MOVPF		PRODL,WREG
		ADDWF		TEMPB2,F
		MOVPF		PRODH,WREG
		ADDWF		TEMPB1,F

D2416UNRMOK
		MOVFP		BARGB0,WREG	; access reciprocal table
		ADDLW		LOW (IBXTBL256)
		MOVWF		TBLPTRL
		MOVLW		HIGH (IBXTBL256)
		CLRF		TBLPTRH,F
		ADDWFC		TBLPTRH,F
		TABLRD		0,1,WREG

D2416ULOOP
		MOVFP		TEMPB0,WREG
		MOVWF		AARGB0
		MOVFP		TEMPB1,WREG
		MOVWF		AARGB1

		CALL		FXD1608U2	; estimate quotient digit

		BTFSS		AARGB0,LSB
		GOTO		D2416UQTEST

		SETF		AARGB1,F
		MOVFP		TEMPB1,WREG
		MOVWF		REMB0
		MOVFP		BARGB0,WREG
		ADDWF		REMB0,F

		BTFSC		_C
		GOTO		D2416UQOK

D2416UQTEST
		MOVFP		AARGB1,WREG	; test
		MULWF		BARGB1

		MOVFP		PRODL,WREG
		SUBWF		TEMPB2,W
		MOVFP		PRODH,WREG
		SUBWFB		REMB0,W

		BTFSC		_C
		GOTO		D2416UQOK

		DECF		AARGB1,F

		MOVFP		BARGB0,WREG
		ADDWF		REMB0,F

		BTFSC		_C
		GOTO		D2416UQOK

		MOVFP		AARGB1,WREG
		MULWF		BARGB1

		MOVFP		PRODL,WREG
		SUBWF		TEMPB2,W
		MOVFP		PRODH,WREG
		SUBWFB		REMB0,W

		BTFSS		_C
		DECF		AARGB1,F

D2416UQOK
		MOVFP		AARGB1,WREG
		MOVWF		ZARGB1

		MOVFP		AARGB1,WREG
		MULWF		BARGB1
		MOVFP		PRODL,WREG
		SUBWF		TEMPB2,F
		MOVFP		PRODH,WREG
		SUBWFB		TEMPB1,F

		MOVFP		AARGB1,WREG
		MULWF		BARGB0
		MOVFP		PRODL,WREG
		SUBWF		TEMPB1,F
		MOVFP		PRODH,WREG
		SUBWFB		TEMPB0,F

		BTFSS		TEMPB0,MSB	; test
		GOTO		D2416QOK
		DECF		ZARGB1,F

		MOVFP		BARGB1,WREG
		ADDWF		TEMPB2,F
		MOVFP		BARGB0,WREG
		ADDWFC		TEMPB1,F

D2416QOK
		DCFSNZ		AEXP,F		; is loop done?
		GOTO		D2416FIXREM

		MOVFP		ZARGB1,WREG
		MOVWF		ZARGB2

		MOVFP		TEMPB1,WREG
		MOVWF		TEMPB0
		MOVFP		TEMPB2,WREG
		MOVWF		TEMPB1
		MOVFP		TEMPB3,WREG
		MOVWF		TEMPB2

		GOTO		D2416ULOOP		

D2416FIXREM
		MOVFP		TEMPB1,WREG
		MOVWF		REMB0
		MOVFP		TEMPB2,WREG
		MOVWF		REMB1

		MOVLW		0x01
		CPFSGT		ZARGB0
		GOTO		D2416REMOK
		RRNCF		ZARGB0,W
		MOVWF		BARGB0
		CALL		DGETNRMD

		MULWF		TEMPB2
		MOVPF		PRODH,REMB1
		MULWF		TEMPB1
		MOVPF		PRODL,WREG
		ADDWF		REMB1,F
		MOVPF		PRODH,REMB0

D2416REMOK
		CLRF		AARGB0,F
		MOVFP		ZARGB1,WREG
		MOVWF		AARGB2
		MOVFP		ZARGB2,WREG
		MOVWF		AARGB1

        RETLW           0x00

		END