;	RCS Header $Id: fxd3224u.asm,v 1.1 2000/03/14 20:32:18 krt Exp $

;	$Revision: 1.1 $

	include <P17C44.INC>		; General 17CXXX definitions

	include <CMATH17.INC>		; Math library definitions

;**********************************************************************************************
;**********************************************************************************************

;       32/24 Bit Unsigned Fixed Point Divide 32/24 -> 32.24

;       Input:  32 bit unsigned fixed point dividend in AARGB0, AARGB1, AARGB2
;               24 bit unsigned fixed point divisor in BARGB0, BARGB1

;       Use:    CALL    FXD3224U

;       Output: 32 bit unsigned fixed point quotient in AARGB0, AARGB1, AARGB2
;               24 bit unsigned fixed point remainder in REMB0, REMB1

;       Result: AARG, REM  <--  AARG / BARG

;	Testing from 100000 trials:

;		min	max	mean
;	Timing:	35	383	261.71	clks

;       PM: 175               DM: 23

;----------------------------------------------------------------------------------------------

PROG	CODE

		EXTERN		FXD3216U
		EXTERN		FXD2424U
		EXTERN		DGETNRMD
		EXTERN		IBXTBL256
		EXTERN		FXD1608U2

FXD3224U
		GLOBAL		FXD3224U

		CLRF		REMB0,F
		CLRF		REMB1,F
		CLRF		REMB2,F
		CLRF		WREG,F
		TSTFSZ		BARGB0
		GOTO		D3224BGT1
		MOVFP		BARGB1,WREG
		MOVWF		BARGB0
		MOVFP		BARGB2,WREG
		MOVWF		BARGB1
		CALL		FXD3216U
		MOVFP		REMB1,WREG
		MOVWF		REMB2
		MOVFP		REMB0,WREG
		MOVWF		REMB1
		CLRF		REMB0,F
		RETLW		0x00

D3224BGT1
		CPFSEQ		AARGB0
		GOTO		D3224AGTB
		MOVFP		AARGB1,WREG
		MOVWF		AARGB0
		MOVFP		AARGB2,WREG
		MOVWF		AARGB1
		MOVFP		AARGB3,WREG
		MOVWF		AARGB2
		CALL		FXD2424U
		
		MOVFP		AARGB2,WREG
		MOVWF		AARGB3
		MOVFP		AARGB1,WREG
		MOVWF		AARGB2
		MOVFP		AARGB0,WREG
		MOVWF		AARGB1
		CLRF		AARGB0,F
		RETLW		0x00
D3224AGTB
		MOVFP		AARGB3,WREG
		MOVWF		BEXP
		MOVFP		AARGB2,WREG
		MOVWF		AARGB3
		MOVFP		AARGB1,WREG
		MOVWF		AARGB2
		MOVFP		AARGB0,WREG
		MOVWF		AARGB1
		CLRF		AARGB0,F

		MOVFP		AARGB0,WREG
		MOVWF		TEMPB0
		MOVFP		AARGB1,WREG
		MOVWF		TEMPB1
		MOVFP		AARGB2,WREG
		MOVWF		TEMPB2
		MOVFP		AARGB3,WREG
		MOVWF		TEMPB3

		MOVLW		0x02		; set loop count
		MOVWF		AEXP

		MOVLW		0x01
		MOVWF		ZARGB0

		BTFSC		BARGB0,MSB
		GOTO		D3224UNRMOK

		CALL		DGETNRMD	; get normalization factor
		MOVWF		ZARGB0

		MULWF		BARGB2
		MOVFP		BARGB1,WREG
		MOVPF		PRODL,BARGB2
		MOVPF		PRODH,BARGB1
		MULWF		ZARGB0
		MOVPF		PRODL,WREG
		ADDWF		BARGB1,F
		MOVFP		BARGB0,WREG
		MOVPF		PRODH,BARGB0
		MULWF		ZARGB0
		MOVFP		PRODL,WREG
		ADDWF		BARGB0,F

		MOVFP		ZARGB0,WREG
		MULWF		AARGB3
		MOVPF		PRODL,TEMPB3
		MOVPF		PRODH,TEMPB2
		MULWF		AARGB1
		MOVPF		PRODL,TEMPB1
		MOVPF		PRODH,TEMPB0
		MULWF		AARGB2
		MOVPF		PRODL,WREG
		ADDWF		TEMPB2,F
		MOVPF		PRODH,WREG
		ADDWF		TEMPB1,F
		MOVFP		ZARGB0,WREG
		MULWF		BEXP
		MOVPF		PRODL,BEXP
		MOVFP		PRODH,WREG
		ADDWF		TEMPB3,F

D3224UNRMOK
		MOVFP		BARGB0,WREG	; access reciprocal table
		ADDLW		LOW (IBXTBL256)
		MOVWF		TBLPTRL
		MOVLW		HIGH (IBXTBL256)
		CLRF		TBLPTRH,F
		ADDWFC		TBLPTRH,F
		TABLRD		0,1,WREG

D3224ULOOP
		MOVFP		TEMPB0,WREG
		MOVWF		AARGB0
		MOVFP		TEMPB1,WREG
		MOVWF		AARGB1

		CALL		FXD1608U2	; estimate quotient digit

		BTFSS		AARGB0,LSB
		GOTO		D3224UQTEST

		SETF		AARGB1,F
		MOVFP		TEMPB1,WREG
		MOVWF		REMB0
		MOVFP		BARGB0,WREG
		ADDWF		REMB0,F

		BTFSC		_C		; test
		GOTO		D3224UQOK

D3224UQTEST
		MOVFP		AARGB1,WREG
		MULWF		BARGB1

		MOVFP		PRODL,WREG
		SUBWF		TEMPB2,W
		MOVFP		PRODH,WREG
		SUBWFB		REMB0,W

		BTFSC		_C		; test
		GOTO		D3224UQOK

		DECF		AARGB1,F

		MOVFP		BARGB0,WREG
		ADDWF		REMB0,F

		BTFSC		_C		; test
		GOTO		D3224UQOK

		MOVFP		AARGB1,WREG
		MULWF		BARGB1

		MOVFP		PRODL,WREG
		SUBWF		TEMPB2,W
		MOVFP		PRODH,WREG
		SUBWFB		REMB0,W

		BTFSS		_C		; test
		DECF		AARGB1,F

D3224UQOK
		MOVFP		AARGB1,WREG
		MOVWF		ZARGB1

		MOVFP		AARGB1,WREG
		MULWF		BARGB2
		MOVFP		PRODL,WREG
		SUBWF		TEMPB3,F
		MOVFP		PRODH,WREG
		SUBWFB		TEMPB2,F
		CLRF		WREG,F
		SUBWFB		TEMPB1,F
		SUBWFB		TEMPB0,F

		MOVFP		AARGB1,WREG
		MULWF		BARGB1
		MOVFP		PRODL,WREG
		SUBWF		TEMPB2,F
		MOVFP		PRODH,WREG
		SUBWFB		TEMPB1,F
		CLRF		WREG,F
		SUBWFB		TEMPB0,F

		MOVFP		AARGB1,WREG
		MULWF		BARGB0
		MOVFP		PRODL,WREG
		SUBWF		TEMPB1,F
		MOVFP		PRODH,WREG
		SUBWFB		TEMPB0,F

		BTFSS		TEMPB0,MSB
		GOTO		D3224QOK
		DECF		ZARGB1,F

		MOVFP		BARGB2,WREG
		ADDWF		TEMPB3,F
		MOVFP		BARGB1,WREG
		ADDWFC		TEMPB2,F
		MOVFP		BARGB0,WREG
		ADDWFC		TEMPB1,F
		CLRF		WREG,F
		ADDWFC		TEMPB0,F

D3224QOK
		DCFSNZ		AEXP,F		; is loop done?
		GOTO		D3224FIXREM

		MOVFP		ZARGB1,WREG
		MOVWF		ZARGB2

		MOVFP		TEMPB1,WREG
		MOVWF		TEMPB0
		MOVFP		TEMPB2,WREG
		MOVWF		TEMPB1
		MOVFP		TEMPB3,WREG
		MOVWF		TEMPB2
		MOVFP		BEXP,WREG
		MOVWF		TEMPB3

		GOTO		D3224ULOOP

D3224FIXREM
		MOVFP		TEMPB1,WREG
		MOVWF		REMB0
		MOVFP		TEMPB2,WREG
		MOVWF		REMB1
		MOVFP		TEMPB3,WREG
		MOVWF		REMB2

		MOVLW		0x01
		CPFSGT		ZARGB0
		GOTO		D3224REMOK
		RRNCF		ZARGB0,W
		MOVWF		BARGB0
		CALL		DGETNRMD

		MOVWF		ZARGB0

		MULWF		TEMPB3
		MOVPF		PRODH,AARGB2
		MULWF		TEMPB2
		MOVPF		PRODL,WREG
		ADDWF		AARGB2,F
		MOVPF		PRODH,REMB1
		MOVFP		ZARGB0,WREG
		MULWF		TEMPB1
		MOVPF		PRODL,WREG
		ADDWF		REMB1,F
		MOVPF		PRODH,REMB0

		MOVFP		AARGB2,WREG
		MOVWF		REMB2

D3224REMOK
		CLRF		AARGB0,F
		CLRF		AARGB1,F
		MOVFP		ZARGB1,WREG
		MOVWF		AARGB3
		MOVFP		ZARGB2,WREG
		MOVWF		AARGB2

        RETLW       0x00

		END
