;	RCS Header $Id: int2416.asm,v 1.1 2000/03/15 19:15:50 krt Exp $

;	$Revision: 1.1 $

	include <P17C44.INC>		; General 17CXXX definitions

	include <CMATH17.INC>		; Math library definitions

;**********************************************************************************************
;**********************************************************************************************

;       Float to integer conversion

;       Input:  24 bit floating point number in AEXP, AARGB0, AARGB1

;       Use:    CALL    INT2416

;       Output: 16 bit 2's complement integer right justified in AARGB0, AARGB1

;       Result: AARG  <--  INT( AARG )

;	Testing on [MINNUM,MAXNUM] from 100000 trials:

;			min		max		mean
;	Timing:	30		59		39.2	clks

;			min		max		mean	rms
;	Error:	0x00	0x00	0.00	0.00	nsb

;       PM: 127							DM: 6

;----------------------------------------------------------------------------------------------

PROG	CODE

		EXTERN		SETIOV2416

INT2416
		GLOBAL		INT2416

		CLRF  		AARGB2,W
		CPFSGT		AEXP		 		; test for zero argument
		RETLW 		0x00
		MOVFP 		AARGB0,WREG
		MOVWF		SIGN	 	 		; save sign in SIGN
		BSF   		AARGB0,MSB	 		; make MSB explicit

		MOVLW 		EXPBIAS+D'15'		; remove bias+15 from EXP
		SUBWF 		AEXP,W

		BTFSS 		WREG,MSB	 		; if >= 15, integer overflow
		GOTO  		SETIOV2416	 		; will occur
		NEGW  		AEXP,F

		MOVLW 		7            		; do byte shift if EXP >= 8
		CPFSGT		AEXP
		GOTO  		SNIB2416
		SUBWF 		AEXP,F		 		; EXP = EXP - 7
		MOVFP 		AARGB1,WREG
		MOVWF		AARGB2		 		; save for rounding
		MOVFP 		AARGB0,WREG
		MOVWF		AARGB1
		CLRF  		AARGB0,F
		DCFSNZ		AEXP,F		 		; EXP = EXP - 1
		GOTO  		SHIFT2416OK  		; shift completed if EXP = 0

		MOVLW 		7            		; do byte shift if EXP >= 8
		CPFSGT		AEXP
		GOTO  		SNIB2416A
		SUBWF 		AEXP,F		 		; EXP = EXP - 7
		MOVFP 		AARGB1,WREG
		MOVWF		AARGB2		 		; save for rounding
		CLRF  		AARGB1,F
		DCFSNZ		AEXP,F		 		; EXP = EXP - 1
		GOTO  		SHIFT2416OK  		; shift completed if EXP = 0

SNIB2416B
		MOVLW 		3            		; do nibble shift if EXP >= 4
		CPFSGT		AEXP
		GOTO  		SHIFT2416B
		SWAPF 		AARGB2,W
		ANDLW 		0x0F
		MOVWF 		AARGB2
		GOTO  		SHIFT2416OK  		; shift completed if EXP = 0

SHIFT2416B
		BCF   		_C           		; at most 3 right shifts are required
		RRCF  		AARGB2,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT2416OK  		; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB2,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT2416OK  		; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB2,F
		GOTO  		SHIFT2416OK

SNIB2416A
		MOVLW 		3			 		; do nibble shift if EXP >= 4
		CPFSGT		AEXP
		GOTO  		SHIFT2416A
		SUBWF 		AEXP,F	   	 		; EXP = EXP - 3
		SWAPF 		AARGB1,W
		MOVWF 		AARGB2		 		; save for rounding
		ANDLW 		0x0F
		MOVWF 		AARGB1
		DCFSNZ		AEXP,F	   	 		; EXP = EXP - 1
		GOTO  		SHIFT2416OK	 		; shift completed if EXP = 0

SHIFT2416A
		BCF   		_C         	 		; at most 3 right shifts are required
		RRCF  		AARGB1,F   	 		; right shift by EXP
		RRCF  		AARGB2,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT2416OK	 		; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT2416OK	 		; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		GOTO  		SHIFT2416OK

SNIB2416
		MOVLW 		3          	 		; do nibble shift if EXP >= 4
		CPFSGT		AEXP
		GOTO  		SHIFT2416
		SUBWF 		AEXP,F	   	 		; EXP = EXP - 3
		SWAPF 		AARGB1,W
		MOVWF 		AARGB2		 		; save for rounding
		ANDLW 		0x0F
		MOVWF 		AARGB1
		SWAPF 		AARGB0,W
		ANDLW 		0xF0
		ADDWF 		AARGB1,F
		SWAPF 		AARGB0,W
		ANDLW 		0x0F
		MOVWF 		AARGB0
		DCFSNZ		AEXP,F	   	 		; EXP = EXP - 1
		GOTO  		SHIFT2416OK	 		; shift completed if EXP = 0

SHIFT2416
		BCF   		_C         	 		; at most 3 right shifts are required
		RRCF  		AARGB0,F   	 		; right shift by EXP
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT2416OK	 		; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB0,F  
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT2416OK	 		; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB0,F  
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F

SHIFT2416OK
		BTFSC 		FPFLAGS,RND	 		; is rounding enabled?
		BTFSS 		AARGB2,MSB 	 		; is NSB > 0x80?
		GOTO  		INT2416OK
		BSF	  		_C		   	 		; set carry for rounding
		MOVLW 		0x80
		CPFSGT		AARGB2	   	 		; if NSB = 0x80, select even
		RRCF  		AARGB1,W   	 		; using lsb in carry
		CLRF  		WREG,F
		ADDWFC		AARGB1,F
		ADDWFC		AARGB0,F
		BTFSC 		AARGB0,MSB
		GOTO  		SETIOV2416

INT2416OK
		BTFSS 		SIGN,MSB     		; if sign bit set, negate               
		RETLW 		0
		COMF  		AARGB1,F
		COMF  		AARGB0,F
		INFSNZ		AARGB1,F
		INCF  		AARGB0,F
		RETLW 		0

		END
