;	RCS Header $Id: int2424.asm,v 1.1 2000/03/15 19:13:55 krt Exp $

;	$Revision: 1.1 $

	include <P17C44.INC>		; General 17CXXX definitions

	include <CMATH17.INC>		; Math library definitions

;**********************************************************************************************
;**********************************************************************************************

;       Float to integer conversion

;       Input:  24 bit floating point number in AEXP, AARGB0, AARGB1

;       Use:    CALL    INT2424

;       Output: 24 bit 2's complement integer right justified in AARGB0, AARGB1, AARGB2

;       Result: AARG  <--  INT( AARG )

;	Testing on [MINNUM,MAXNUM] from 100000 trials:

;			min		max		mean
;	Timing:	33		72		44.1		clks

;			min		max		mean	rms
;	Error:	0x00	0x00	0.00	0.00	nsb

;       PM: 185							DM: 7

;----------------------------------------------------------------------------------------------

PROG	CODE

		EXTERN		SETIOV2424

INT2424
		GLOBAL		INT2424

		CLRF  		AARGB2,W
		CPFSGT		AEXP		 		; test for zero argument
		RETLW 		0x00
		MOVFP 		AARGB0,WREG
		MOVWF		SIGN	 	 		; save sign in SIGN
		BSF   		AARGB0,MSB	 		; make MSB explicit

		CLRF  		AARGB3,F

		MOVLW 		EXPBIAS+D'23'		; remove bias+23 from EXP
		SUBWF 		AEXP,W

		BTFSS 		WREG,MSB	 		; if >= 23, integer overflow
		GOTO  		SETIOV2424	 		; will occur
		NEGW  		AEXP,F

		MOVLW 		7            		; do byte shift if EXP >= 8
		CPFSGT		AEXP
		GOTO  		SNIB2424
		SUBWF 		AEXP,F		 		; EXP = EXP - 7
		MOVFP 		AARGB2,WREG
		MOVWF		AARGB3		 		; save for rounding
		MOVFP 		AARGB1,WREG
		MOVWF		AARGB2
		MOVFP 		AARGB0,WREG
		MOVWF		AARGB1
		CLRF  		AARGB0,F
		DCFSNZ		AEXP,F		 		; EXP = EXP - 1
		GOTO  		SHIFT2424OK  		; shift completed if EXP = 0

		MOVLW 		7            		; do byte shift if EXP >= 8
		CPFSGT		AEXP         		; do another byte shift if EXP >= 8
		GOTO  		SNIB2424A
		SUBWF 		AEXP,F		 		; EXP = EXP - 7
		MOVFP 		AARGB2,WREG
		MOVWF		AARGB3		 		; save for rounding
		MOVFP 		AARGB1,WREG
		MOVWF		AARGB2
		CLRF  		AARGB1,F
		DCFSNZ		AEXP,F		 		; EXP = EXP - 1
		GOTO  		SHIFT2424OK  		; shift completed if EXP = 0

		MOVLW 		7            		; do byte shift if EXP >= 8
		CPFSGT		AEXP         		; do another byte shift if EXP >= 8
		GOTO  		SNIB2424B
		SUBWF 		AEXP,F		 		; EXP = EXP - 7
		MOVFP 		AARGB2,WREG
		MOVWF		AARGB3		 		; save for rounding
		CLRF  		AARGB2,F
		DCFSNZ		AEXP,F		 		; EXP = EXP - 1
		GOTO  		SHIFT2424OK  		; shift completed if EXP = 0

SNIB2424C
		MOVLW 		3          	 		; do nibble shift if EXP >= 4
		CPFSGT		AEXP
		GOTO  		SHIFT2424C
		SWAPF 		AARGB3,W
		ANDLW 		0x0F
		MOVWF 		AARGB3
		GOTO  		SHIFT2424OK	 		; shift completed if EXP = 0

SHIFT2424C
		BCF   		_C         	 		; at most 3 right shifts are required
		RRCF  		AARGB3,F   	 		; right shift by EXP
		DCFSNZ		AEXP,F
		GOTO  		SHIFT2424OK	 		; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB3,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT2424OK	 		; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB3,F
		GOTO  		SHIFT2424OK

SNIB2424B
		MOVLW 		3          	 		; do nibble shift if EXP >= 4
		CPFSGT		AEXP
		GOTO  		SHIFT2424B
		SUBWF 		AEXP,F	   	 		; EXP = EXP - 3
		SWAPF 		AARGB2,W
		MOVWF 		AARGB3		 		; save for rounding
		ANDLW 		0x0F
		MOVWF 		AARGB2
		DCFSNZ		AEXP,F	   	 		; EXP = EXP - 1
		GOTO  		SHIFT2424OK	 		; shift completed if EXP = 0

SHIFT2424B
		BCF   		_C         	 		; at most 3 right shifts are required
		RRCF  		AARGB2,F   	 		; right shift by EXP
		RRCF  		AARGB3,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT2424OK	 		; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT2424OK	 		; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		GOTO  		SHIFT2424OK

SNIB2424A
		MOVLW 		3          	 		; do nibble shift if EXP >= 4
		CPFSGT		AEXP
		GOTO  		SHIFT2424A
		SUBWF 		AEXP,F	   	 		; EXP = EXP - 3
		SWAPF 		AARGB2,W
		MOVWF 		AARGB3		 		; save for rounding
		ANDLW 		0x0F
		MOVWF 		AARGB2

		SWAPF 		AARGB1,W
		ANDLW 		0xF0
		ADDWF 		AARGB2,F

		SWAPF 		AARGB1,W
		ANDLW 		0x0F
		MOVWF 		AARGB1
		DCFSNZ		AEXP,F	   	 		; EXP = EXP - 1
		GOTO  		SHIFT2424OK	 		; shift completed if EXP = 0

SHIFT2424A
		BCF   		_C         	 		; at most 3 right shifts are required
		RRCF  		AARGB1,F   	 		; right shift by EXP
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT2424OK	 		; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT2424OK	 		; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		GOTO  		SHIFT2424OK

SNIB2424
		MOVLW 		3		   	 		; do nibble shift if EXP >= 4
		CPFSGT		AEXP
		GOTO  		SHIFT2424
		SUBWF 		AEXP,F	   	 		; EXP = EXP - 3
		SWAPF 		AARGB2,W
		MOVWF 		AARGB3		 		; save for rounding
		ANDLW 		0x0F
		MOVWF 		AARGB2

		SWAPF 		AARGB1,W
		ANDLW 		0xF0
		ADDWF 		AARGB2,F

		SWAPF 		AARGB1,W
		ANDLW 		0x0F
		MOVWF 		AARGB1

		SWAPF 		AARGB0,W
		ANDLW 		0xF0
		ADDWF 		AARGB1,F

		SWAPF 		AARGB0,W
		ANDLW 		0x0F
		MOVWF 		AARGB0
		DCFSNZ		AEXP,F		 		; EXP = EXP - 1
		GOTO  		SHIFT2424OK  		; shift completed if EXP = 0

SHIFT2424
		BCF   		_C         	 		; at most 3 right shifts are required
		RRCF  		AARGB0,F   	 		; right shift by EXP
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT2424OK	 		; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB0,F
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F
		DCFSNZ		AEXP,F
		GOTO  		SHIFT2424OK	 		; shift completed if EXP = 0
		BCF   		_C
		RRCF  		AARGB0,F  
		RRCF  		AARGB1,F
		RRCF  		AARGB2,F
		RRCF  		AARGB3,F

SHIFT2424OK
		BTFSC 		FPFLAGS,RND	 		; is rounding enabled?
		BTFSS 		AARGB3,MSB 	 		; is NSB > 0x80?
		GOTO  		INT2424OK
		BSF	  		_C		   	 		; set carry for rounding
		MOVLW 		0x80
		CPFSGT		AARGB3	   	 		; if NSB = 0x80, select even
		RRCF  		AARGB2,W   	 		; using lsb in carry
		CLRF  		WREG,F
		ADDWFC		AARGB2,F
		ADDWFC		AARGB1,F
		ADDWFC		AARGB0,F
		BTFSC 		AARGB0,MSB
		GOTO  		SETIOV2424

INT2424OK
		BTFSS 		SIGN,MSB	 		; if sign bit set, negate               
		RETLW 		0
		COMF  		AARGB2,F
		COMF  		AARGB1,F
		COMF  		AARGB0,F
		INCF  		AARGB2,F
		CLRF  		WREG,F
		ADDWFC		AARGB1,F
		ADDWFC		AARGB0,F
		RETLW 		0

		END
